from __future__ import annotations
import logging

from angr.analyses import AnalysesHub
from . import Analysis

l = logging.getLogger(name=__name__)


class SootClassHierarchyError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class NoConcreteDispatch(SootClassHierarchyError):
    def __init__(self, msg):
        self.msg = msg


class SootClassHierarchy(Analysis):
    """
    Generate complete hierarchy.
    """

    def __init__(self):
        if self.project.arch.name != "Soot":
            raise SootClassHierarchyError("SootClassHierarchyError only supports analyzing Soot programs.")

        self.interface_implementers = {}
        self.sub_interfaces = {}
        self.dir_sub_interfaces = {}
        self.sub_classes = {}
        self.dir_sub_classes = {}
        # init data
        self.init_hierarchy()

    def init_hierarchy(self):
        for _class_name, cls in self.project.loader.main_object.classes.items():
            if "INTERFACE" in cls.attrs:
                self.interface_implementers[cls] = []
                self.dir_sub_interfaces[cls] = []
            else:
                self.dir_sub_classes[cls] = []

        for _class_name, cls in self.project.loader.main_object.classes.items():
            if self.has_super_class(cls):
                if "INTERFACE" in cls.attrs:
                    # TODO
                    # super_interfaces
                    pass

                else:
                    super_class = self.project.loader.main_object.classes[cls.super_class]
                    self.dir_sub_classes[super_class].append(cls)

                    for i_name in cls.interfaces:
                        # get interface
                        if i_name not in self.project.loader.main_object.classes:
                            continue
                        i = self.project.loader.main_object.classes[i_name]
                        self.interface_implementers[i].append(cls)

        # fill direct implementers with subclasses
        for _class_name, cls in self.project.loader.main_object.classes.items():
            if "INTERFACE" in cls.attrs:
                implementers = self.interface_implementers[cls]
                s = set()

                for c in implementers:
                    s |= set(self.get_sub_classes_including(c))

                self.interface_implementers[cls] = list(s)

    def has_super_class(self, cls):
        if cls.super_class:
            try:
                self.project.loader.main_object.classes[cls.super_class]
                return True
            except KeyError:
                pass

        return False

    def is_subclass_including(self, cls_child, cls_parent):
        parent_classes = self.get_super_classes_including(cls_child)

        return cls_parent in parent_classes

    def is_subclass(self, cls_child, cls_parent):
        parent_classes = self.get_super_classes(cls_child)

        return cls_parent in parent_classes

    def is_visible_method(self, cls, method):
        method_cls = self.project.loader.main_object.classes[method.class_name]

        if not self.is_visible_class(cls, method_cls):
            return False

        if "PUBLIC" in method.attrs:
            return True

        if "PRIVATE" in method.attrs:
            return cls == method_cls

        package_from = cls.name.split(".")[:-1]
        package_to = method_cls.name.split(".")[:-1]

        if "PROTECTED" in method.attrs:
            is_sub = self.is_subclass_including(cls, method_cls)
            is_same_package = package_from == package_to
            return is_sub or is_same_package

        return package_from == package_to

    def is_visible_class(self, cls_from, cls_to):
        if "PUBLIC" in cls_to.attrs:
            return True

        if "PROTECTED" in cls_to.attrs or "PRIVATE" in cls_to.attrs:
            return False

        package_from = cls_from.name.split(".")[:-1]
        package_to = cls_to.name.split(".")[:-1]
        return package_from == package_to

    def get_super_classes(self, cls):
        if "INTERFACE" in cls.attrs:
            raise SootClassHierarchyError("This is an Interface")

        super_classes = []

        current = cls
        try:
            while True:
                current = self.project.loader.main_object.classes[current.super_class]
                super_classes.append(current)

        except KeyError:
            return super_classes

    def get_super_classes_including(self, cls):
        super_classes = self.get_super_classes(cls)
        res = []

        res.append(cls)
        res.extend(super_classes)

        return res

    def get_implementers(self, interface):
        if "INTERFACE" not in interface.attrs:
            raise SootClassHierarchyError("This is not an interface")

        res_set = set()

        for i in self.get_sub_interfaces_including(interface):
            res_set |= set(self.interface_implementers[i])

        return list(res_set)

    def get_sub_interfaces_including(self, interface):
        res = self.get_sub_interfaces(interface)
        res.append(interface)

        return res

    def get_sub_interfaces(self, interface):
        if "INTERFACE" not in interface.attrs:
            raise SootClassHierarchyError("This is not an interface")

        if interface in self.sub_interfaces:
            return self.sub_interfaces[interface]

        # Otherwise
        res = []
        for i in self.dir_sub_interfaces[interface]:
            res.extend(self.get_sub_interfaces_including(i))

        self.sub_interfaces[interface] = res
        return res

    def get_sub_classes(self, cls):
        if "INTERFACE" in cls.attrs:
            raise SootClassHierarchyError("This is an Interface. Class needed")

        if cls in self.sub_classes:
            return self.sub_classes[cls]

        res = []
        for c in self.dir_sub_classes[cls]:
            res.extend(self.get_sub_classes_including(c))

        self.sub_classes[cls] = res
        return res

    def get_sub_classes_including(self, cls):
        if "INTERFACE" in cls.attrs:
            raise SootClassHierarchyError("This is an Interface. Class needed")

        res = []
        res.extend(self.get_sub_classes(cls))
        res.append(cls)

        return res

    def resolve_abstract_dispatch(self, cls, method):
        if "INTERFACE" in cls.attrs:
            classes_set = set()
            for i in self.get_implementers(cls):
                classes_set |= set(self.get_sub_classes_including(i))
            classes = list(classes_set)
        else:
            classes = self.get_sub_classes_including(cls)

        res_set = set()
        for c in classes:
            if "ABSTRACT" not in c.attrs:
                res_set.add(self.resolve_concrete_dispatch(c, method))

        return list(res_set)

    def resolve_concrete_dispatch(self, cls, method):
        if "INTERFACE" in cls.attrs:
            raise SootClassHierarchyError("class needed!")

        for c in self.get_super_classes_including(cls):
            for m in c.methods:
                if m.name == method.name and m.params == method.params and self.is_visible_method(c, method):
                    return m

        raise NoConcreteDispatch("Could not resolve concrete dispatch!")

    def resolve_special_dispatch(self, method, container):
        # container is the method that contains the invoke
        method_cls = self.project.loader.main_object.classes[method.class_name]
        container_cls = self.project.loader.main_object.classes[container.class_name]

        if method.name == "<init>" or "PRIVATE" in method.attrs:
            return method

        if self.is_subclass(method_cls, container_cls):
            return self.resolve_concrete_dispatch(container_cls, method)

        return method

    def resolve_invoke(self, invoke_expr, method, container):
        # Generic method to resolve invoke
        # Given an invoke expression it figures out which "technique" should be applied
        invoke_type = str(type(invoke_expr))
        cls = self.project.loader.main_object.classes[method.class_name]

        if (
            "VirtualInvokeExpr" in invoke_type
            or "DynamicInvokeExpr" in invoke_type
            or "InterfaceInvokeExpr" in invoke_type
        ):
            targets = self.resolve_abstract_dispatch(cls, method)

        elif "SpecialInvokeExpr" in invoke_type:
            t = self.resolve_special_dispatch(method, container)
            targets = [t]

        elif "StaticInvokeExpr" in invoke_type:
            targets = [method]

        return targets


AnalysesHub.register_default("SootClassHierarchy", SootClassHierarchy)



from __future__ import annotations
from itertools import chain
from typing import TYPE_CHECKING, Any, TypeVar
from collections.abc import Iterable, MutableMapping
from collections.abc import Callable
from functools import partial
import logging

import networkx

from angr import ailment
from angr.ailment import Block
from angr.ailment.expression import Phi, VirtualVariable, VirtualVariableCategory
from angr.ailment.statement import Assignment, Label, Statement

from angr.analyses.decompiler.ailgraph_walker import traverse_in_order
from angr.code_location import AILCodeLocation
from angr.knowledge_plugins.functions.function import Function
from angr.utils.ail import is_head_controlled_loop_block
from angr.utils.ssa import is_phi_assignment
from .rewriting_engine import SimEngineSSARewriting
from .rewriting_state import RewritingState

if TYPE_CHECKING:
    from angr.analyses.decompiler.ssailification.ssailification import Def, UDef
    from angr.project import Project

l = logging.getLogger(__name__)
T = TypeVar("T")
U = TypeVar("U")


class RewritingAnalysis:
    """
    RewritingAnalysis traverses the AIL graph, inserts phi nodes, and rewrites all expression uses to virtual variables
    when necessary.
    """

    def __init__(
        self,
        project: Project,
        func: Function,
        ail_graph: networkx.DiGraph[Block],
        phiid_to_udef: dict[int, UDef],
        block_to_phiids: dict[Block, list[int]],
        rewrite_tmps: bool,
        ail_manager,
        func_args: set[VirtualVariable],
        def_to_udef: MutableMapping[Def, UDef],
        extern_defs: set[UDef],
        incomplete_defs: set[Def],
        vvar_id_start: int = 0,
        stackvars: bool = False,
        fail_fast: bool = False,
    ):
        self.project = project
        self._fail_fast = fail_fast
        self._function = func

        self._graph = ail_graph
        self._phiid_to_udef = phiid_to_udef
        self._block_to_phiids = block_to_phiids
        self._rewrite_tmps = rewrite_tmps
        self._ail_manager = ail_manager
        self._func_args = func_args
        self._extern_defs = extern_defs
        self._engine_ail = SimEngineSSARewriting(
            self.project,
            rewrite_tmps=self._rewrite_tmps,
            incomplete_defs=incomplete_defs,
            ail_manager=ail_manager,
            vvar_id_start=vvar_id_start,
            def_to_udef=def_to_udef,
            stackvars=stackvars,
            fail_fast=self._fail_fast,
        )

        self._pending_states: dict[ailment.Block, RewritingState] = {}
        self.out_blocks = {}
        self.out_states = {}
        self.resized_func_args: dict[VirtualVariable, VirtualVariable] = {}
        # loop_states stores states at the beginning of a loop block *after a loop iteration*, where the block is the
        # following:
        #    0x4036df | t4 = (rcx<8> == 0x0<64>)
        #    0x4036df | if (t4) { Goto 0x4036e2<64> } else { Goto 0x4036df<64> }
        #    0x4036df | STORE(addr=t3, data=t2, size=8, endness=Iend_LE, guard=None)
        #    0x4036df | rdi<8> = t8
        #
        self.head_controlled_loop_outstates = {}

        self._analyze()

        self.out_graph = self._make_new_graph(ail_graph)

    def _analyze(self):
        self._pre_analysis()
        entry_block = next((n for n in self._graph if n.addr == self._function.addr), None)
        entry_blocks = {n for n in self._graph if not self._graph.pred[n]}
        if entry_block is not None:
            entry_blocks.add(entry_block)
        traverse_in_order(self._graph, sorted(entry_blocks), self._run_on_node)
        self._post_analysis()

    @property
    def max_vvar_id(self) -> int | None:
        return self._engine_ail.current_vvar_id

    def _make_new_graph(self, old_graph: networkx.DiGraph[ailment.Block]) -> networkx.DiGraph[ailment.Block]:
        new_graph: networkx.DiGraph[ailment.Block] = networkx.DiGraph()
        for node in old_graph:
            new_graph.add_node(self.out_blocks.get((node.addr, node.idx), node))

        for src, dst, data in old_graph.edges(data=True):
            new_graph.add_edge(
                self.out_blocks.get((src.addr, src.idx), src),
                self.out_blocks.get((dst.addr, dst.idx), dst),
                **data,
            )

        return new_graph

    def create_phi_statements(
        self, node: Block, phiid_to_udef: dict[int, UDef], phi_ids: Iterable[int]
    ) -> list[Assignment]:
        phi_stmts: list[Assignment] = []
        seen: set[tuple[str, int]] = set()
        for phi_id in phi_ids:
            udef = phiid_to_udef[phi_id]
            category = udef[0]
            for suboffset in range(udef[1], udef[1] + udef[2]):
                key = (udef[0], suboffset)
                assert not self._fail_fast or key not in seen, "Overlapping phis in a single block"
                seen.add(key)

            match category:
                case "reg":
                    _, reg_offset, reg_bytes = udef

                    phi_var = Phi(
                        self._ail_manager.next_atom(),
                        reg_bytes * self.project.arch.byte_width,
                        src_and_vvars=[],  # back patch later
                        ins_addr=node.addr,
                    )
                    phi_dst = VirtualVariable(
                        self._ail_manager.next_atom(),
                        self._engine_ail._current_vvar_id,
                        reg_bytes * self.project.arch.byte_width,
                        VirtualVariableCategory.REGISTER,
                        oident=reg_offset,
                        ins_addr=node.addr,
                    )
                    self._engine_ail._current_vvar_id += 1

                case "stack":
                    _, stack_offset, stack_size = udef

                    phi_var = Phi(
                        self._ail_manager.next_atom(),
                        stack_size * self.project.arch.byte_width,
                        src_and_vvars=[],  # back patch later
                        ins_addr=node.addr,
                    )
                    phi_dst = VirtualVariable(
                        self._ail_manager.next_atom(),
                        self._engine_ail._current_vvar_id,
                        stack_size * self.project.arch.byte_width,
                        VirtualVariableCategory.STACK,
                        oident=stack_offset,
                        ins_addr=node.addr,
                    )
                    self._engine_ail._current_vvar_id += 1
                case _:
                    raise NotImplementedError

            phi_stmt = Assignment(
                None,
                phi_dst,
                phi_var,
                ins_addr=node.addr,
            )
            phi_stmts.append(phi_stmt)

        return phi_stmts

    @staticmethod
    def insert_phi_statements(node: Block, phi_stmts: list[Assignment]):
        idx = 0
        while idx < len(node.statements):
            if not isinstance(node.statements[idx], Label):
                break
            idx += 1

        if idx >= len(node.statements):
            node.statements += phi_stmts
        else:
            node.statements = node.statements[:idx] + phi_stmts + node.statements[idx:]

    def _reg_predicate(self, node_: Block, *, reg_offset: int) -> tuple[bool, Any]:
        out_state: RewritingState = (
            self.head_controlled_loop_outstates[(node_.addr, node_.idx)]
            if is_head_controlled_loop_block(node_) and (node_.addr, node_.idx) in self.head_controlled_loop_outstates
            else self.out_states[(node_.addr, node_.idx)]
        )
        if reg_offset in out_state.registers:
            # we found a perfect hit
            existing_var = out_state.registers[reg_offset]
            if existing_var is None:
                # the vvar is not set. it should never be referenced
                return True, None
            vvar = existing_var.deep_copy(self._ail_manager)
            return True, vvar

        return False, None

    def _stack_predicate(self, node_: Block, *, stack_offset: int) -> tuple[bool, Any]:
        out_state: RewritingState = (
            self.head_controlled_loop_outstates[(node_.addr, node_.idx)]
            if is_head_controlled_loop_block(node_)
            else self.out_states[(node_.addr, node_.idx)]
        )
        if stack_offset in out_state.stackvars:
            existing_var = out_state.stackvars[stack_offset]
            if existing_var is None:
                # the vvar is not set. it should never be referenced
                return True, None
            vvar = existing_var.deep_copy(self._ail_manager)
            return True, vvar
        return False, None

    #
    # Main analysis routines
    #

    def _pre_analysis(self):
        for node in self._graph:
            phi_stmts = self.create_phi_statements(node, self._phiid_to_udef, self._block_to_phiids[node])
            if phi_stmts:
                self.insert_phi_statements(node, phi_stmts)

    def _initial_abstract_state(self, node) -> RewritingState:
        func_args_map = {}
        # we may identify args that are not identified by Traversal. make sure these still get used
        unused_func_args = set(self._func_args)
        for arg_vvar in self._func_args:
            offset = (
                arg_vvar.parameter_reg_offset
                if arg_vvar.parameter_category == VirtualVariableCategory.REGISTER
                else (
                    arg_vvar.parameter_stack_offset
                    if arg_vvar.parameter_category == VirtualVariableCategory.STACK
                    else (
                        arg_vvar.reg_offset
                        if arg_vvar.category == VirtualVariableCategory.REGISTER
                        else arg_vvar.stack_offset
                        if arg_vvar.category == VirtualVariableCategory.STACK
                        else None
                    )
                )
            )
            assert offset is not None
            for suboffset in range(offset, offset + arg_vvar.size):
                func_args_map[(arg_vvar.parameter_category, suboffset)] = arg_vvar

        state = RewritingState(
            AILCodeLocation(node.addr, node.idx, 0, node.addr),
            self.project.arch,
            self._function,
            node,
        )
        more_args: list[VirtualVariable] = []
        for kind, offset, size in sorted(self._extern_defs):
            category = VirtualVariableCategory.REGISTER if kind == "reg" else VirtualVariableCategory.STACK
            if (arg_vvar := func_args_map.get((category, offset))) is not None:
                unused_func_args.discard(arg_vvar)
                if arg_vvar.size == size:
                    vvar = arg_vvar
                elif arg_vvar in self.resized_func_args:
                    vvar = self.resized_func_args[arg_vvar]
                else:
                    # rewriting the size the arg
                    vvar = VirtualVariable(
                        arg_vvar.idx,
                        arg_vvar.varid,
                        size * 8,
                        VirtualVariableCategory.PARAMETER,
                        (category, offset),
                        **arg_vvar.tags,
                    )
                    self.resized_func_args[arg_vvar] = vvar
            else:
                varid = self._engine_ail._current_vvar_id
                self._engine_ail._current_vvar_id += 1
                vvar = VirtualVariable(self._engine_ail.ail_manager.next_atom(), varid, size * 8, category, offset)
            more_args.append(vvar)

        # update state with function arguments
        for func_arg in chain(more_args, unused_func_args):
            if func_arg.category == VirtualVariableCategory.REGISTER:
                reg_offset = func_arg.reg_offset
                for suboff in range(reg_offset, reg_offset + func_arg.size):
                    state.registers[suboff] = func_arg
            elif func_arg.category == VirtualVariableCategory.STACK:
                stack_offset = func_arg.stack_offset
                for suboff in range(stack_offset, stack_offset + func_arg.size):
                    state.stackvars[suboff] = func_arg
            elif func_arg.parameter_category == VirtualVariableCategory.REGISTER:
                reg_offset = func_arg.parameter_reg_offset
                assert reg_offset is not None
                for suboff in range(reg_offset, reg_offset + func_arg.size):
                    state.registers[suboff] = func_arg
            elif func_arg.parameter_category == VirtualVariableCategory.STACK:
                stack_offset = func_arg.parameter_stack_offset
                assert stack_offset is not None
                for suboff in range(stack_offset, stack_offset + func_arg.size):
                    state.stackvars[suboff] = func_arg

        return state

    def _run_on_node(self, node: Block):
        """

        :param node:    The current node.
        :param state:   The analysis state.
        :return:        A tuple: (any changes occur, successor state)
        """

        state = self._pending_states.get(node, None)
        state = self._initial_abstract_state(node) if state is None else state.copy()
        block_key = (node.addr, node.idx)

        state.loc = AILCodeLocation(node.addr, node.idx, 0, node.addr)

        self._engine_ail.process(
            state,
            block=node,
        )

        if self._engine_ail.hclb_side_exit_state is not None:
            self.head_controlled_loop_outstates[block_key] = state
            state = self._engine_ail.hclb_side_exit_state
            self._engine_ail.hclb_side_exit_state = None
        self.out_states[block_key] = state
        self.out_blocks[block_key] = self._engine_ail.out_block
        for succ in self._graph.succ[node]:
            self._pending_states[succ] = state

    def _post_analysis(self):
        # update phi nodes
        for original_node in self._graph:
            node_key = original_node.addr, original_node.idx
            node = self.out_blocks.get(node_key, original_node)

            if any(
                isinstance(stmt, Assignment) and isinstance(stmt.dst, VirtualVariable) and isinstance(stmt.src, Phi)
                for stmt in node.statements
            ):
                # there we go
                phi_stmts = []
                phi_extraction_stmts = []
                other_stmts = []
                for stmt in node.statements:
                    if (
                        isinstance(stmt, Assignment)
                        and isinstance(stmt.dst, VirtualVariable)
                        and isinstance(stmt.src, Phi)
                        and len(stmt.src.src_and_vvars) > 0
                    ):
                        # avoid re-assignment
                        phi_stmts.append(stmt)
                        continue
                    if not is_phi_assignment(stmt):
                        other_stmts.append(stmt)
                        continue
                    assert isinstance(stmt, Assignment)
                    assert isinstance(stmt.dst, VirtualVariable)
                    assert isinstance(stmt.src, Phi)

                    src_and_vvars = []
                    if stmt.dst.was_reg:
                        perfect_matches = []
                        for pred in self._graph.predecessors(original_node):
                            vvar = self._follow_one_path_backward(
                                self._graph,
                                pred,
                                partial(self._reg_predicate, reg_offset=stmt.dst.reg_offset),
                            )
                            src_and_vvars.append(((pred.addr, pred.idx), vvar))
                            perfect_matches.append(
                                (vvar.reg_offset == stmt.dst.reg_offset and vvar.size == stmt.dst.size)
                                if vvar is not None
                                else True
                            )
                        if all(perfect_matches):
                            phi_var = Phi(stmt.src.idx, stmt.src.bits, src_and_vvars=src_and_vvars)
                            phi_stmt = Assignment(stmt.idx, stmt.dst, phi_var, **stmt.tags)
                            phi_stmts.append(phi_stmt)
                        else:
                            # different sizes of vvars found; this means the variable is unused from this point on
                            continue

                    elif stmt.dst.was_stack:
                        for pred in self._graph.predecessors(original_node):
                            vvar = self._follow_one_path_backward(
                                self._graph,
                                pred,
                                partial(
                                    self._stack_predicate,
                                    stack_offset=stmt.dst.stack_offset,
                                ),
                            )
                            src_and_vvars.append(((pred.addr, pred.idx), vvar))

                        phi_var = Phi(stmt.src.idx, stmt.src.bits, src_and_vvars=src_and_vvars)
                        phi_stmt = Assignment(stmt.idx, stmt.dst, phi_var, **stmt.tags)
                        phi_stmts.append(phi_stmt)
                    else:
                        raise NotImplementedError

                node = node.copy(statements=phi_stmts + phi_extraction_stmts + other_stmts)
                self.out_blocks[node_key] = node

    @staticmethod
    def _follow_one_path_backward(
        graph: networkx.DiGraph[T], src: T, predicate: Callable[[T], tuple[bool, U]]
    ) -> U | None:
        visited = set()
        return_value = None
        the_node = src
        while the_node not in visited:
            visited.add(the_node)
            stop, return_value = predicate(the_node)
            if stop:
                break
            # keep going
            more_preds = list(graph.predecessors(the_node))
            if len(more_preds) != 1:
                # no longer a single path back
                break
            the_node = more_preds[0]
        return return_value

    @staticmethod
    def _find_phi_vvar_for_reg(reg_offset: int, reg_size: int, stmts: list[Statement]) -> VirtualVariable | None:
        for stmt in stmts:
            if (
                isinstance(stmt, Assignment)
                and isinstance(stmt.src, Phi)
                and isinstance(stmt.dst, VirtualVariable)
                and stmt.dst.was_reg
                and stmt.dst.reg_offset == reg_offset
                and stmt.dst.size == reg_size
            ):
                return stmt.dst
        return None



from __future__ import annotations
from typing import TYPE_CHECKING
import logging

import claripy
import pyvex

from angr.knowledge_plugins.propagations import PropagationModel
from angr.utils.constants import DEFAULT_STATEMENT
from angr.code_location import CodeLocation
from angr.blade import Blade
from angr.analyses.propagator import vex_vars
from angr.utils.vex import get_tmp_def_stmt
from .resolver import IndirectJumpResolver
from .propagator_utils import PropagatorLoadCallback

if TYPE_CHECKING:
    from angr import Block

_l = logging.getLogger(name=__name__)


def exists_in_replacements(replacements, block_loc, tmp_var):
    exists = False
    for rep in replacements:
        if rep == block_loc:
            exists = True
            break

    if not exists:
        return False

    exists = False
    for var in replacements[block_loc]:
        if var == tmp_var:
            exists = True
            break

    return exists


class ConstantResolver(IndirectJumpResolver):
    """
    Resolve an indirect jump by running a constant propagation on the entire function and check if the indirect jump can
    be resolved to a constant value. This resolver must be run after all other more specific resolvers.
    """

    def __init__(self, project, max_func_nodes: int = 512):
        super().__init__(project, timeless=False)
        self.max_func_nodes = max_func_nodes

        # stats
        self._resolved = 0
        self._unresolved = 0
        self._cache_hits = 0
        self._props_saved = 0

    def filter(self, cfg, addr, func_addr, block, jumpkind):
        if not cfg.functions.contains_addr(func_addr):
            # the function does not exist
            return False

        # for performance, we don't run constant resolver if the function is too large
        func_block_count = cfg.functions.get_func_block_count(func_addr)
        if func_block_count is None or func_block_count > self.max_func_nodes:
            return False

        # we support both an indirect call and jump since the value can be resolved

        return jumpkind in {"Ijk_Boring", "Ijk_Call"}

    def resolve(  # pylint:disable=unused-argument
        self, cfg, addr: int, func_addr: int, block: Block, jumpkind: str, func_graph_complete: bool = True, **kwargs
    ):
        """
        This function does the actual resolve. Our process is easy:
        Propagate all values inside the function specified, then extract
        the tmp_var used for the indirect jump from the basic block.
        Use the tmp var to locate the constant value stored in the replacements.
        If not present, returns False tuple.

        :param cfg:         CFG with specified function
        :param addr:        Address of indirect jump
        :param func_addr:   Address of function of indirect jump
        :param block:       Block of indirect jump (Block object)
        :param jumpkind:    VEX jumpkind (Ijk_Boring or Ijk_Call)
        :return:            Bool tuple with replacement address
        """
        if not cfg.functions.contains_addr(func_addr):
            #  the function does not exist
            return False, []

        func = cfg.functions.get_by_addr(func_addr)

        vex_block = block.vex
        if isinstance(vex_block.next, pyvex.expr.RdTmp):
            tmp_stmt_idx, tmp_ins_addr = self._find_tmp_write_stmt_and_ins(vex_block, vex_block.next.tmp)
            if tmp_stmt_idx is None or tmp_ins_addr is None:
                return False, []

            # first check: is it jumping to a target loaded from memory? if so, it should have been resolved by
            # MemoryLoadResolver.
            stmt = vex_block.statements[tmp_stmt_idx]
            assert isinstance(stmt, pyvex.IRStmt.WrTmp)
            if (
                isinstance(stmt.data, pyvex.IRExpr.Load)
                and isinstance(stmt.data.addr, pyvex.IRExpr.Const)
                and stmt.data.result_size(vex_block.tyenv) == self.project.arch.bits
            ):
                # well, if MemoryLoadResolver hasn't resolved it, we can try to resolve it here, or bail early because
                # ConstantResolver won't help.
                load_addr = stmt.data.addr.con.value
                try:
                    value = self.project.loader.memory.unpack_word(load_addr, size=self.project.arch.bytes)
                    if isinstance(value, int) and self._is_target_valid(cfg, value):
                        return True, [value]
                except KeyError:
                    pass
                return False, []

            # second check: what does the jump rely on? slice it back and see
            b = Blade(
                cfg.graph,
                addr,
                -1,
                cfg=cfg,
                project=self.project,
                ignore_bp=False,
                ignore_sp=False,
                max_level=3,
                stop_at_calls=True,
                cross_insn_opt=True,
                control_dependence=False,
            )
            stmt_loc = addr, DEFAULT_STATEMENT
            if self._check_jump_target_is_loaded_from_dynamic_addr(b, stmt_loc):
                # loading from memory - unsupported
                return False, []
            if self._check_jump_target_is_compared_against(b, stmt_loc):
                # the jump/call target is compared against another value, which means it's not deterministic
                # ConstantResolver does not support such cases by design
                return False, []

            # first check the replacements cache
            resolved_tmp = None
            is_full_func_prop = None
            block_loc = CodeLocation(block.addr, tmp_stmt_idx, ins_addr=tmp_ins_addr)
            tmp_var = vex_vars.VEXTmp(vex_block.next.tmp)
            prop_key = "FCP", func_addr
            cached_prop = cfg.kb.propagations.get(prop_key)
            if cached_prop is not None:
                is_full_func_prop = len(func.block_addrs_set) == cached_prop.function_block_count
                replacements = cached_prop.replacements
                if exists_in_replacements(replacements, block_loc, tmp_var):
                    self._cache_hits += 1
                    resolved_tmp = replacements[block_loc][tmp_var]

            if resolved_tmp is None and is_full_func_prop:
                self._props_saved += 1

            if resolved_tmp is None and not is_full_func_prop:
                _l.debug("ConstantResolver: Propagating for %r at %#x.", func, addr)
                prop = self.project.analyses.FastConstantPropagation(
                    func,
                    vex_cross_insn_opt=False,
                    load_callback=PropagatorLoadCallback(self.project).propagator_load_callback,
                )
                # update the cache
                model = PropagationModel(
                    prop_key, replacements=prop.replacements, function_block_count=len(func.block_addrs_set)
                )
                cfg.kb.propagations.update(prop_key, model)

                replacements = prop.replacements
                if replacements and exists_in_replacements(replacements, block_loc, tmp_var):
                    resolved_tmp = replacements[block_loc][tmp_var]

            if resolved_tmp is not None:
                if (
                    isinstance(resolved_tmp, claripy.ast.Base)
                    and resolved_tmp.op == "BVV"
                    and self._is_target_valid(cfg, resolved_tmp.args[0])
                ):
                    self._resolved += 1
                    # print(f"{self._resolved} ({self._props_saved} saved, {self._cache_hits} cached) / "
                    #       f"{self._resolved + self._unresolved}")
                    # print(f"+ Function: {func_addr:#x}, block {addr:#x}, target {resolved_tmp.args[0]:#x}")
                    return True, [resolved_tmp.args[0]]
                if isinstance(resolved_tmp, int) and self._is_target_valid(cfg, resolved_tmp):
                    self._resolved += 1
                    # print(f"{self._resolved} ({self._props_saved} saved, {self._cache_hits} cached) / "
                    #       f"{self._resolved + self._unresolved}")
                    # print(f"+ Function: {func_addr:#x}, block {addr:#x}, target {resolved_tmp:#x}")
                    return True, [resolved_tmp]

        self._unresolved += 1
        # print(f"{RESOLVED} ({SAVED_PROPS} saved, {HIT_CACHE} cached) / {RESOLVED + UNRESOLVED}")
        # print(f"- Function: {func_addr:#x}, block {addr:#x}, FAILED")
        return False, []

    def _check_jump_target_is_loaded_from_dynamic_addr(self, b, stmt_loc) -> bool:
        queue: list[tuple[int, int, int]] = []  # depth, block_addr, stmt_idx
        seen_locs: set[tuple[int, int]] = set()
        for block_addr, stmt_idx in b.slice.predecessors(stmt_loc):
            if (block_addr, stmt_idx) in seen_locs:
                continue
            seen_locs.add((block_addr, stmt_idx))
            queue.append((0, block_addr, stmt_idx))
        while queue:
            depth, pred_addr, stmt_idx = queue.pop(0)
            if depth >= 3:
                break

            # skip all IMarks
            if stmt_idx != DEFAULT_STATEMENT:
                block = self.project.factory.block(pred_addr, cross_insn_opt=True).vex
                stmt = block.statements[stmt_idx]
                if isinstance(stmt, pyvex.IRStmt.IMark):
                    for succ_addr, succ_stmt_idx in b.slice.predecessors((pred_addr, stmt_idx)):
                        if (succ_addr, succ_stmt_idx) in seen_locs:
                            continue
                        seen_locs.add((succ_addr, succ_stmt_idx))
                        queue.append((depth + 1 if succ_addr != pred_addr else depth, succ_addr, succ_stmt_idx))
                    continue

                if (
                    isinstance(stmt, pyvex.IRStmt.WrTmp)
                    and isinstance(stmt.data, pyvex.IRExpr.Load)
                    and not isinstance(stmt.data.addr, pyvex.IRExpr.Const)
                ):
                    # loading from memory
                    return True

            for succ_addr, succ_stmt_idx in b.slice.predecessors((pred_addr, stmt_idx)):
                if (succ_addr, succ_stmt_idx) in seen_locs:
                    continue
                seen_locs.add((succ_addr, succ_stmt_idx))
                queue.append((depth + 1 if succ_addr != pred_addr else depth, succ_addr, succ_stmt_idx))

        return False

    def _check_jump_target_is_compared_against(self, b, stmt_loc) -> bool:
        # let's find which register the jump uses
        jump_site = self.project.factory.block(stmt_loc[0], cross_insn_opt=True).vex
        if not isinstance(jump_site.next, pyvex.IRExpr.RdTmp):
            return False
        next_tmp = jump_site.next.tmp
        # find its definition
        next_tmp_def = get_tmp_def_stmt(jump_site, next_tmp)
        if next_tmp_def is None:
            return False
        next_tmp_def_stmt = jump_site.statements[next_tmp_def]
        if not (
            isinstance(next_tmp_def_stmt, pyvex.IRStmt.WrTmp) and isinstance(next_tmp_def_stmt.data, pyvex.IRExpr.Get)
        ):
            return False
        next_reg = next_tmp_def_stmt.data.offset

        # traverse back at most one level and check:
        # - this register has never been updated
        # - a comparison is conducted on this register (via a tmp, most likely)
        queue = []
        seen = set()
        for block_addr, stmt_idx in b.slice.predecessors(stmt_loc):
            if (block_addr, stmt_idx) in seen:
                continue
            seen.add((block_addr, stmt_idx))
            queue.append((0, block_addr, stmt_idx))
        while queue:
            depth, pred_addr, stmt_idx = queue.pop(0)
            if depth > 1:
                continue

            # skip all IMarks
            pred = pred_addr, stmt_idx
            if stmt_idx != DEFAULT_STATEMENT:
                block = self.project.factory.block(pred_addr, cross_insn_opt=True).vex
                stmt = block.statements[stmt_idx]
                if isinstance(stmt, pyvex.IRStmt.IMark):
                    for succ_addr, succ_stmt_idx in b.slice.predecessors(pred):
                        if (succ_addr, succ_stmt_idx) in seen:
                            continue
                        seen.add((succ_addr, succ_stmt_idx))
                        queue.append((depth + 1 if succ_addr != pred_addr else depth, succ_addr, succ_stmt_idx))
                    continue

                if isinstance(stmt, pyvex.IRStmt.Put) and stmt.offset == next_reg:
                    # this register has been updated before we find a comparison; do not continue along this path
                    continue

                if (
                    isinstance(stmt, pyvex.IRStmt.WrTmp)
                    and isinstance(stmt.data, pyvex.IRExpr.Binop)
                    and stmt.data.op.startswith("Iop_Cmp")
                ):
                    # what is it comparing against?
                    for arg in stmt.data.args:
                        if isinstance(arg, pyvex.IRExpr.RdTmp):
                            arg_tmp_def = get_tmp_def_stmt(block, arg.tmp)
                            if arg_tmp_def is not None:
                                arg_tmp_def_stmt = block.statements[arg_tmp_def]
                                if (
                                    isinstance(arg_tmp_def_stmt, pyvex.IRStmt.WrTmp)
                                    and isinstance(arg_tmp_def_stmt.data, pyvex.IRExpr.Get)
                                    and arg_tmp_def_stmt.data.offset == next_reg
                                ):
                                    # the jump target is compared against this register
                                    return True
                                # another case: VEX optimization may have caused the tmp to be stored in the target
                                # register. we need handle this case as well.
                                if any(
                                    isinstance(stmt_, pyvex.IRStmt.Put)
                                    and stmt_.offset == next_reg
                                    and isinstance(stmt_.data, pyvex.IRExpr.RdTmp)
                                    and stmt_.data.tmp == arg.tmp
                                    for stmt_ in block.statements[arg_tmp_def + 1 : stmt_idx]
                                ):
                                    # the jump target is compared against this register
                                    return True

            # continue traversing predecessors
            for succ_addr, succ_stmt_idx in b.slice.predecessors(pred):
                if (succ_addr, succ_stmt_idx) in seen:
                    continue
                seen.add((succ_addr, succ_stmt_idx))
                queue.append((depth + 1 if succ_addr != pred_addr else depth, succ_addr, succ_stmt_idx))

        return False

    @staticmethod
    def _find_tmp_write_stmt_and_ins(vex_block, tmp: int) -> tuple[int | None, int | None]:
        stmt_idx = None
        for idx, stmt in enumerate(reversed(vex_block.statements)):
            if isinstance(stmt, pyvex.IRStmt.IMark) and stmt_idx is not None:
                ins_addr = stmt.addr + stmt.delta
                return stmt_idx, ins_addr
            if isinstance(stmt, pyvex.IRStmt.WrTmp) and stmt.tmp == tmp:
                stmt_idx = len(vex_block.statements) - idx - 1
        return None, None



# pylint:disable=line-too-long
from __future__ import annotations
import logging
import math

from angr.ailment import Expr

from .engine_base import SimplifierAILEngine, SimplifierAILState
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class DivSimplifierAILEngine(SimplifierAILEngine):
    """
    An AIL pass for the div simplifier
    """

    @staticmethod
    def _check_divisor(a: int, b: int, ndigits: int = 6) -> int | None:
        if b == 0:
            return None
        divisor_1 = 1 + (a // b)
        divisor_2 = int(round(a / float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None

    # pylint: disable=too-many-boolean-expressions
    def _handle_expr_Convert(self, expr: Expr.Convert):
        if expr.from_bits == 128 and expr.to_bits == 64:
            operand_expr = self._expr(expr.operand)
            if (
                (
                    isinstance(operand_expr, Expr.BinaryOp)
                    and operand_expr.op == "Mul"
                    and isinstance(operand_expr.operands[1], Expr.Const)
                    and isinstance(operand_expr.operands[0], Expr.BinaryOp)
                )
                and operand_expr.operands[0].op in {"Shr", "Mod"}
                and isinstance(operand_expr.operands[0].operands[1], Expr.Const)
            ):
                if operand_expr.operands[0].op == "Shr":
                    Y = operand_expr.operands[0].operands[1].value
                    assert isinstance(Y, int)
                else:
                    Y = int(math.log2(operand_expr.operands[0].operands[1].value))
                C = operand_expr.operands[1].value
                assert isinstance(C, int)
                divisor = self._check_divisor(pow(2, 64 + Y), C)
                if divisor:
                    X = operand_expr.operands[0].operands[0]
                    new_const = Expr.Const(expr.idx, None, divisor, 64)
                    return Expr.BinaryOp(expr.idx, "Div", [X, new_const], expr.signed, **expr.tags)

        return expr

    def _handle_binop_Shr(self, expr):
        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        X = None
        divisor = None

        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.BinaryOp)
            and operand_0.op == "Mod"
            and isinstance(operand_0.operands[1], Expr.Const)
        ):
            divisor = operand_0.operands[1].value * pow(2, operand_1.value)
            X = operand_0.operands[0]
        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.Convert)
            and isinstance(operand_0.operand, Expr.BinaryOp)
            and operand_0.operand.op == "Mod"
            and isinstance(operand_0.operand.operands[1], Expr.Const)
        ):
            divisor = operand_0.operand.operands[1].value * pow(2, operand_1.value)
            X = operand_0.operand.operands[0]
        if (
            (
                isinstance(operand_1, Expr.Const)
                and isinstance(operand_0, Expr.Convert)
                and operand_0.from_bits == 128
                and operand_0.to_bits == 64
            )
            and isinstance(operand_0.operand, Expr.BinaryOp)
            and operand_0.operand.op == "Mul"
        ):
            if isinstance(operand_0.operand.operands[1], Expr.Const):
                C = operand_0.operand.operands[1].value
                Y = operand_1.value
                assert isinstance(C, int)
                assert isinstance(Y, int)
                divisor = self._check_divisor(pow(2, 64 + Y), C)
                X = operand_0.operand.operands[0]
            elif isinstance(operand_0.operand.operands[0], Expr.BinaryOp) and operand_0.operand.operands[0].op in {
                "Shr",
                "Mod",
            }:
                C = operand_0.operand.operands[1].value
                Z = operand_1.value
                if operand_0.operand.operands[0].op == "Shr":
                    Y = operand_0.operand.operands[0].operands[1].value
                else:
                    Y = int(math.log2(operand_0.operand.operands[0].operands[1].value))
                assert isinstance(C, int)
                assert isinstance(Y, int)
                assert isinstance(Z, int)
                divisor = self._check_divisor(pow(2, 64 + Z + Y), C)
                X = operand_0.operand.operands[0].operands[0]
        if isinstance(operand_1, Expr.Const) and isinstance(operand_0, Expr.BinaryOp) and operand_0.op == "Add":
            add_0, add_1 = operand_0.operands
            Z = operand_1.value
            if add_0.has_atom(add_1) or add_1.has_atom(add_0):
                xC = add_1 if add_0.has_atom(add_1) else add_0
                x_xC = add_0 if add_0.has_atom(add_1) else add_1
                if isinstance(xC, Expr.Convert) and (xC.from_bits > xC.to_bits):
                    Y = xC.from_bits - xC.to_bits
                    if isinstance(xC.operand, Expr.BinaryOp) and xC.operand.op == "Mul":
                        xC_ = xC.operand
                        if isinstance(xC_.operands[1], Expr.Const):
                            C = xC_.operands[1].value
                            X = xC_.operands[0]
                            if isinstance(x_xC, Expr.BinaryOp) and x_xC.op == "Shr":
                                V_, V = x_xC.operands
                                if isinstance(V, Expr.Const):
                                    V = V.value
                                    if (
                                        isinstance(V_, Expr.BinaryOp)
                                        and V_.op == "Sub"
                                        and V_.operands[0] == X
                                        and V_.operands[1] == xC
                                    ):
                                        assert isinstance(Y, int)
                                        assert isinstance(Z, int)
                                        assert isinstance(V, int)
                                        divisor = self._check_divisor(
                                            pow(2, Y + V + Z), C * (pow(2, V) - 1) + pow(2, Y)
                                        )
                # unsigned int here
                if isinstance(xC, Expr.BinaryOp) and xC.op == "Mul":
                    if isinstance(xC.operands[1], Expr.Const) and isinstance(xC.operands[0], Expr.Convert):
                        C = xC.operands[1].value
                        X = xC.operands[0]
                        Y = X.from_bits - X.to_bits
                        if isinstance(x_xC, Expr.BinaryOp) and x_xC.op == "Shr":
                            V_, V = x_xC.operands
                            if isinstance(V, Expr.Const):
                                V = V.value
                                if isinstance(V_, Expr.BinaryOp) and V_.op == "Sub" and V_.operands[1] == xC:
                                    assert isinstance(Y, int)
                                    assert isinstance(Z, int)
                                    assert isinstance(V, int)
                                    divisor = self._check_divisor(pow(2, Y + V + Z), C * (pow(2, V) - 1) + pow(2, Y))
                elif (
                    isinstance(xC, Expr.BinaryOp)
                    and xC.op == "Shr"
                    and (
                        isinstance(xC.operands[1], Expr.Const)
                        and isinstance(xC.operands[0], Expr.BinaryOp)
                        and xC.operands[0].op == "Mul"
                        and isinstance(xC.operands[0].operands[1], Expr.Const)
                    )
                    and (
                        isinstance(x_xC, Expr.BinaryOp)
                        and isinstance(x_xC.operands[1], Expr.Const)
                        and isinstance(x_xC.operands[0], Expr.BinaryOp)
                        and x_xC.op == "Shr"
                        and x_xC.operands[0].op == "Sub"
                    )
                ):
                    X = xC.operands[0].operands[0]
                    C = xC.operands[0].operands[1].value
                    Y = xC.operands[1].value
                    V = x_xC.operands[1].value
                    if x_xC.operands[0].operands[0] == X:
                        assert isinstance(Y, int)
                        assert isinstance(Z, int)
                        assert isinstance(V, int)
                        divisor = self._check_divisor(pow(2, Y + V + Z), C * (pow(2, V) - 1) + pow(2, Y))

        # unsigned int
        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.BinaryOp)
            and operand_0.op == "Mul"
            and isinstance(operand_0.operands[1], Expr.Const)
        ):
            if isinstance(operand_0.operands[0], Expr.Convert):
                V = operand_0.operands[0].from_bits - operand_0.operands[0].to_bits
                C = operand_0.operands[1].value
                Z = operand_1.value
                X = operand_0.operands[0]
                assert isinstance(C, int)
                assert isinstance(Z, int)
                assert isinstance(V, int)
                divisor = self._check_divisor(pow(2, V + Z), C)
            elif (
                isinstance(operand_0.operands[0], Expr.BinaryOp)
                and isinstance(operand_0.operands[0].operands[1], Expr.Const)
                and operand_0.operands[0].op in {"Shr", "Mod"}
            ):
                X = operand_0.operands[0].operands[0]
                V = 0
                ndigits = 6
                if isinstance(X, Expr.Convert):
                    V = X.from_bits - X.to_bits
                if V == 32:
                    ndigits = 5
                C = operand_0.operands[1].value
                Y = operand_0.operands[0].operands[1].value
                if operand_0.operands[0].op == "Mod":
                    Y = int(math.log2(operand_0.operands[0].operands[1].value))
                Z = operand_1.value
                assert isinstance(Y, int)
                assert isinstance(Z, int)
                assert isinstance(V, int)
                assert isinstance(C, int)
                divisor = self._check_divisor(pow(2, Y + Z + V), C, ndigits)
            else:
                X = operand_0.operands[0]
                Y = operand_1.value
                C = operand_0.operands[1].value
                assert isinstance(Y, int)
                assert isinstance(C, int)
                divisor = self._check_divisor(pow(2, Y), C)

        if divisor and X:
            new_const = Expr.Const(expr.idx, None, divisor, 64)
            return Expr.BinaryOp(expr.idx, "Div", [X, new_const], expr.signed, **expr.tags)

        if isinstance(operand_1, Expr.Const):
            if isinstance(operand_0, Expr.VirtualVariable) and operand_0.was_reg:
                new_operand = Expr.Const(operand_1.idx, None, 2**operand_1.value, operand_0.bits)
                return Expr.BinaryOp(expr.idx, "Div", [operand_0, new_operand], expr.signed)
            if (
                isinstance(operand_0, Expr.BinaryOp)
                and operand_0.op == "Shr"
                and isinstance(operand_0.operands[1], Expr.Const)
            ):
                new_const = Expr.Const(
                    operand_1.idx, None, operand_0.operands[1].value + operand_1.value, operand_1.bits
                )
                return Expr.BinaryOp(expr.idx, "Shr", [operand_0.operands[0], new_const], expr.signed, **expr.tags)

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, "Shr", [operand_0, operand_1], expr.signed)
        return expr

    def _handle_binop_Mul(self, expr):
        operand_0, operand_1 = expr.operands

        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.BinaryOp)
            and isinstance(operand_0.operands[1], Expr.Const)
            and operand_0.op in {"Mod", "Shr"}
        ):
            assert isinstance(operand_0.operands[1].value, int)
            assert isinstance(operand_1.value, int)
            Y = int(math.log2(operand_0.operands[1].value)) if operand_0.op == "Mod" else operand_0.operands[1].value
            C = operand_1.value
            X = operand_0.operands[0]
            # there is a convert outside this expr
            V = 64
            if isinstance(X, Expr.Convert):
                V = X.from_bits - X.to_bits
            ndigits = 5 if V == 32 else 6
            if (divisor := self._check_divisor(pow(2, V + Y), C, ndigits)) and X:
                new_const = Expr.Const(expr.idx, None, divisor, 64)
                return Expr.BinaryOp(expr.idx, "Div", [X, new_const], expr.signed, **expr.tags)
        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.Convert)
            and isinstance(operand_0.operand, Expr.BinaryOp)
            and isinstance(operand_0.operand.operands[1], Expr.Const)
            and operand_0.operand.op in {"Mod", "Shr"}
            and isinstance(operand_1.value, int)
        ):
            assert isinstance(operand_0.operand.operands[1].value, int)
            if operand_0.operand.op == "Mod":
                Y = int(math.log2(operand_0.operand.operands[1].value))
            else:
                Y = operand_0.operand.operands[1].value
            C = operand_1.value
            X = operand_0.operand.operands[0]
            V = operand_0.from_bits - operand_0.to_bits
            ndigits = 5 if V == 32 else 6
            if (divisor := self._check_divisor(pow(2, V + Y), C, ndigits)) and X:
                new_const = Expr.Const(expr.idx, None, divisor, 64)
                return Expr.BinaryOp(expr.idx, "Div", [X, new_const], expr.signed, **expr.tags)
        return expr

    def _handle_binop_Div(self, expr):
        operand_0 = self._expr(expr.operands[0])
        operand_1 = self._expr(expr.operands[1])

        if (
            isinstance(operand_1, Expr.Const)
            and isinstance(operand_0, Expr.BinaryOp)
            and operand_0.op in {"Div", "Mod"}
            and isinstance(operand_0.operands[1], Expr.Const)
        ):
            new_const_value = operand_1.value * operand_0.operands[1].value
            new_const = Expr.Const(operand_1.idx, None, new_const_value, operand_1.bits)
            return Expr.BinaryOp(expr.idx, "Div", [operand_0.operands[0], new_const], expr.signed, **expr.tags)

        if (operand_0, operand_1) != (expr.operands[0], expr.operands[1]):
            return Expr.BinaryOp(expr.idx, "Div", [operand_0, operand_1], expr.signed, **expr.tags)
        return expr

    def _handle_binop_Add(self, expr):
        op0 = self._expr(expr.operands[0])
        op1 = self._expr(expr.operands[1])

        matched, new_expr = self._match_signed_division_add_operands(op0, op1)
        if matched:
            return new_expr
        matched, new_expr = self._match_signed_division_add_operands(op1, op0)  # pylint:disable=arguments-out-of-order
        if matched:
            return new_expr

        return expr

    def _match_signed_division_add_operands(self, op0, op1):
        # From: Add((Conv(64->32, ((Load(addr=stack_base+4, size=4, endness=Iend_LE) Mulls 0x55555556<32>)
        #            >> 0x20<8>)) >> 0x1f<8>),
        #            Conv(64->32, ((Load(addr=stack_base+4, size=4, endness=Iend_LE) Mulls 0x55555556<32>) >> 0x20<8>)))
        # To: Load(addr=stack_base+4, size=4, endness=Iend_LE) /s 3

        # op0
        if not (
            isinstance(op0, Expr.BinaryOp)
            and op0.op == "Shr"
            and isinstance(op0.operands[1], Expr.Const)
            and op0.operands[1].value == 0x1F
        ):
            return False, None
        if not (
            isinstance(op0.operands[0], Expr.Convert)
            and op0.operands[0].from_bits == 64
            and op0.operands[0].to_bits == 32
        ):
            return False, None

        op0_inner = op0.operands[0].operand
        if not (
            isinstance(op0_inner, Expr.BinaryOp)
            and op0_inner.op == "Shr"
            and isinstance(op0_inner.operands[1], Expr.Const)
            and op0_inner.operands[1].value == 32
        ):
            return False, None

        # op1
        if op1 != op0.operands[0]:
            return False, None

        # extract
        inner = op0_inner.operands[0]
        if isinstance(inner, Expr.BinaryOp) and inner.op == "Mull" and inner.signed:
            operand_0, operand_1 = inner.operands

            if isinstance(operand_1, Expr.Const) and not isinstance(operand_0, Expr.Const):
                # swap them
                operand_0, operand_1 = operand_1, operand_0

            if isinstance(operand_0, Expr.Const) and not isinstance(operand_1, Expr.Const) and operand_0.bits == 32:
                bits = operand_0.bits
                C = operand_0.value
                X = operand_1
                V = bits
                ndigits = 5 if V == 32 else 6
                assert isinstance(C, int)
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None and X:
                    new_const = Expr.Const(None, None, divisor, V)
                    new_expr = Expr.BinaryOp(inner.idx, "Div", [X, new_const], inner.signed, **inner.tags)
                    return True, new_expr

        return False, None


class DivSimplifier(OptimizationPass):
    """
    Simplifies various division optimizations back to "div".
    """

    ARCHES = [
        "X86",
        "AMD64",
        "ARMCortexM",
        "ARMHF",
        "ARMEL",
    ]
    PLATFORMS = None  # everything
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Simplify arithmetic division"
    DESCRIPTION = __doc__.strip()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.state = SimplifierAILState(self.project.arch)
        self.engine = DivSimplifierAILEngine(self.project)

        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        for block in list(self._graph.nodes()):
            new_block = block
            old_block = None

            while new_block != old_block:
                old_block = new_block
                new_block = self.engine.process(state=self.state.copy(), block=old_block.copy())
                _l.debug("new block: %s", new_block.statements)

            self._update_block(block, new_block)



from __future__ import annotations

from .base import (
    BaseStructuredCodeGenerator,
    InstructionMapping,
    InstructionMappingElement,
    PositionMappingElement,
    PositionMapping,
)
from .c import CStructuredCodeGenerator, CStructuredCodeWalker
from .dwarf_import import ImportSourceCode
from .dummy import DummyStructuredCodeGenerator

__all__ = (
    "BaseStructuredCodeGenerator",
    "CStructuredCodeGenerator",
    "CStructuredCodeWalker",
    "DummyStructuredCodeGenerator",
    "ImportSourceCode",
    "InstructionMapping",
    "InstructionMappingElement",
    "PositionMapping",
    "PositionMappingElement",
)



from __future__ import annotations
from collections import defaultdict
import logging
from typing import Any, TYPE_CHECKING

from angr.ailment.block import Block
from angr.ailment.statement import Assignment
from angr.ailment.expression import VirtualVariable, Phi, BinaryOp, UnaryOp

import angr
from angr.utils.ail import is_phi_assignment
from angr.knowledge_plugins.functions import Function
from angr.analyses import register_analysis
from angr.analyses.decompiler.structuring.structurer_nodes import SequenceNode, LoopNode
from angr.analyses.decompiler.sequence_walker import SequenceWalker
from .dephication_base import DephicationBase
from .rewriting_engine import SimEngineDephiRewriting

if TYPE_CHECKING:
    from angr import KnowledgeBase


l = logging.getLogger(__name__)


class PhiAssignmentCollector(SequenceWalker):
    """
    PhiAssignmentCollector collects all phi variables and their corresponding source virtual variables in a
    SequenceNode.
    """

    def __init__(self, seq_node: SequenceNode):
        super().__init__(
            handlers={
                Block: self._handle_Block,
            }
        )

        self.phi_to_src: dict[int, set[int]] = defaultdict(set)

        self.walk(seq_node)

    def _handle_Block(self, block: Block, **kwargs) -> None:  # pylint:disable=unused-argument
        for stmt in block.statements:
            if isinstance(stmt, Assignment) and isinstance(stmt.dst, VirtualVariable) and isinstance(stmt.src, Phi):
                for _, vvar in stmt.src.src_and_vvars:
                    if vvar is None:
                        l.debug("Invalid vvar None found in %r.src.src_and_vvars.", stmt)
                    else:
                        self.phi_to_src[stmt.dst.varid].add(vvar.varid)


class SeqNodeRewriter(SequenceWalker):
    """
    SeqNodeRewriter rewrites a SequenceNode by replacing all phi variables with their corresponding source virtual
    variables.
    """

    def __init__(
        self,
        seq_node: SequenceNode,
        vvar_to_vvar: dict[int, int],
        project: angr.Project,
        variable_kb: KnowledgeBase | None = None,
        func_addr: int | None = None,
    ):
        super().__init__(
            handlers={
                Block: self._handle_Block,
                # statement handlers
                Assignment: self._handle_Assignment,
                # expression handlers
                BinaryOp: self._handle_BinaryOp,
                UnaryOp: self._handle_UnaryOp,
            }
        )

        self.vvar_to_vvar = vvar_to_vvar
        self.engine = SimEngineDephiRewriting(project, self.vvar_to_vvar, func_addr=func_addr, variable_kb=variable_kb)

        self.output = self.walk(seq_node)
        if self.output is None:
            # nothing is changed during rewriting
            self.output = seq_node

    def _handle_Assignment(self, stmt: Assignment, **kwargs) -> Assignment:  # pylint:disable=unused-argument
        return self.engine._handle_stmt_Assignment(stmt)

    def _handle_BinaryOp(self, expr, **kwargs):  # pylint:disable=unused-argument
        return self.engine._handle_expr_BinaryOp(expr)

    def _handle_UnaryOp(self, expr, **kwargs):  # pylint:disable=unused-argument
        return self.engine._handle_expr_UnaryOp(expr)

    def _handle_Block(self, block: Block, **kwargs) -> Block | None:  # pylint:disable=unused-argument
        self.engine.out_block = None
        self.engine.process(None, block=block)
        if self.engine.out_block is not None:
            out = self.engine.out_block
            out.statements = [stmt for stmt in out.statements if not is_phi_assignment(stmt)]
            self.engine.out_block = None
            return out
        return None

    def _handle_Loop(self, node: LoopNode, **kwargs):
        new_loop = super()._handle_Loop(node, **kwargs)
        changed = False
        if new_loop is None:
            new_loop = node
        else:
            changed = True

        if is_phi_assignment(new_loop.initializer):
            changed = True
            new_loop.initializer = None
        return new_loop if changed else None


class SeqNodeDephication(DephicationBase):
    """
    SeqNodeDephication removes phi expressions from a SequenceNode and its children. It also removes redundant variable
    assignments, e.g., `vvar_2 = vvar_1` where both vvar_1 and vvar_2 are mapped to the same variable.
    """

    def __init__(
        self,
        func: Function | str,
        seq_node,
        vvar_to_vvar_mapping: dict[int, int] | None = None,
        rewrite: bool = False,
        variable_kb: KnowledgeBase | None = None,
    ):
        super().__init__(func, vvar_to_vvar_mapping=vvar_to_vvar_mapping, rewrite=rewrite, variable_kb=variable_kb)

        self._seq_node = seq_node

        self._analyze()

    def _collect_phi_assignments(self) -> dict[int, set[int]]:
        # traverse children of the SequenceNode object and find all phi assignment statements
        collector = PhiAssignmentCollector(self._seq_node)
        return collector.phi_to_src

    def _rewrite_container(self) -> Any:
        rewriter = SeqNodeRewriter(
            self._seq_node,
            self.vvar_to_vvar_mapping,
            self.project,
            func_addr=self._function.addr,
            variable_kb=self.variable_kb,
        )
        return rewriter.output


register_analysis(SeqNodeDephication, "SeqNodeDephication")



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Semantic variable naming for array indices.

This module detects variables used as array indices and names them appropriately.
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from angr.ailment import Block
from angr.ailment.expression import BinaryOp, Const, Load
from angr.ailment.statement import Store
from angr.sim_variable import SimVariable

from .naming_base import ClinicNamingBase

if TYPE_CHECKING:
    from angr.ailment import Expression


l = logging.getLogger(name=__name__)

# Names for array indices (used when not already named as loop counter)
ARRAY_INDEX_NAMES = ["idx", "index", "pos", "off"]


class ArrayIndexNaming(ClinicNamingBase):
    """
    Detects variables used as array indices and renames them.

    Array index patterns detected:
    - base + idx * scale (common array indexing)
    - base + idx (byte array or pointer arithmetic)
    - ptr[idx] style access
    """

    PRIORITY = 50  # Lower priority than loop counters

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._index_vars: dict[SimVariable, int] = {}  # var -> usage count
        self._name_counter = 0

    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze the graph for array index patterns.

        :return: Dictionary mapping SimVariable to new name
        """
        if not self._graph:
            return {}

        # Find all variables used as array indices
        self._find_array_indices()

        # Assign names based on usage frequency
        self._assign_index_names()

        return self._var_to_new_name

    def _find_array_indices(self) -> None:
        """
        Find variables used in array indexing patterns.
        """
        for node in self._graph:
            if not isinstance(node, Block):
                continue

            for stmt in node.statements:
                # Check Load expressions
                self._check_expr_for_indices(stmt)

    def _check_expr_for_indices(self, node) -> None:
        """
        Recursively check an expression/statement for array index patterns.
        """
        if node is None:
            return

        # Check if this is a Load with array-style addressing
        if isinstance(node, Load):
            self._analyze_address_expr(node.addr)
            # Also check nested expressions
            self._check_expr_for_indices(node.addr)
            return

        # Check Store statements
        if isinstance(node, Store):
            self._analyze_address_expr(node.addr)
            self._check_expr_for_indices(node.addr)
            self._check_expr_for_indices(node.data)
            return

        # Recursively check BinaryOp operands
        if isinstance(node, BinaryOp):
            for operand in node.operands:
                self._check_expr_for_indices(operand)
            return

        # Check statement attributes
        if hasattr(node, "src"):
            self._check_expr_for_indices(node.src)
        if hasattr(node, "dst"):
            self._check_expr_for_indices(node.dst)

    def _analyze_address_expr(self, addr_expr: Expression) -> None:
        """
        Analyze an address expression for array indexing patterns.

        Patterns:
        - base + idx * scale  (e.g., arr + i * 4)
        - base + idx          (e.g., arr + i for byte arrays)
        - (base + idx * scale) for nested expressions
        """
        if not isinstance(addr_expr, BinaryOp):
            return

        if addr_expr.op != "Add":
            return

        op0, op1 = addr_expr.operands

        # Pattern: base + idx * scale
        if isinstance(op1, BinaryOp) and op1.op == "Mul":
            self._check_mul_for_index(op1)
        elif isinstance(op0, BinaryOp) and op0.op == "Mul":
            self._check_mul_for_index(op0)

        # Pattern: base + idx (could be byte array or already scaled)
        # Only consider if one operand looks like a base (constant or large value)
        # and the other is a variable
        else:
            self._check_add_for_index(op0, op1)

    def _check_mul_for_index(self, mul_expr: BinaryOp) -> None:
        """
        Check a multiplication expression for array index pattern (idx * scale).
        """
        op0, op1 = mul_expr.operands

        # One operand should be a constant (the scale factor)
        # The other should be a variable (the index)
        if isinstance(op1, Const):
            # op0 is potentially the index
            var = self._get_linked_variable(op0)
            if var is not None:
                self._record_index_var(var)
        elif isinstance(op0, Const):
            # op1 is potentially the index
            var = self._get_linked_variable(op1)
            if var is not None:
                self._record_index_var(var)

    def _check_add_for_index(self, op0, op1) -> None:
        """
        Check an addition for simple array index pattern (base + idx).
        """

        # Heuristic: if one operand is a constant (base address) and the other is a variable, the variable might be an
        # index
        var0 = self._get_linked_variable(op0)
        var1 = self._get_linked_variable(op1)

        if isinstance(op0, Const) and var1 is not None:
            # op1 is the potential index
            self._record_index_var(var1)
        elif isinstance(op1, Const) and var0 is not None:
            # op0 is the potential index
            self._record_index_var(var0)

    def _record_index_var(self, var: SimVariable) -> None:
        """
        Record a variable as being used as an array index.
        """
        if var not in self._index_vars:
            self._index_vars[var] = 0
        self._index_vars[var] += 1

    def _assign_index_names(self) -> None:
        """
        Assign names to array index variables.
        """

        # Sort by usage count (most used first)
        sorted_vars = sorted(self._index_vars.items(), key=lambda x: -x[1])

        for var, _count in sorted_vars:
            # Assign a name
            if self._name_counter < len(ARRAY_INDEX_NAMES):
                new_name = ARRAY_INDEX_NAMES[self._name_counter]
            else:
                new_name = f"idx{self._name_counter}"

            self._var_to_new_name[var] = new_name
            self._name_counter += 1



from __future__ import annotations

from angr.errors import SimSegfaultException
from angr import sim_options as o
from .paged_memory_mixin import PagedMemoryMixin


class PrivilegedPagingMixin(PagedMemoryMixin):
    """
    A mixin for paged memory models which will raise SimSegfaultExceptions if STRICT_PAGE_ACCESS is enabled and
    a segfault condition is detected.

    Segfault conditions include:
    - getting a page for reading which is non-readable
    - getting a page for writing which is non-writable
    - creating a page

    The latter condition means that this should be inserted under any mixins which provide other implementations of
    ``_initialize_page``.
    """

    def _get_page(self, pageno: int, writing: bool, priv: bool = False, **kwargs):
        page = super()._get_page(pageno, writing, **kwargs)
        if self.category == "mem" and not priv and o.STRICT_PAGE_ACCESS in self.state.options:
            if writing and not self.state.solver.is_true(page.perm_write):
                raise SimSegfaultException(pageno * self.page_size, "non-writable")
            if not writing and not self.state.solver.is_true(page.perm_read):
                raise SimSegfaultException(pageno * self.page_size, "non-readable")

        return page

    def _initialize_page(self, pageno: int, *, priv: bool = False, **kwargs):
        if self.category == "mem" and not priv and o.STRICT_PAGE_ACCESS in self.state.options:
            raise SimSegfaultException(pageno * self.page_size, "unmapped")

        return super()._initialize_page(pageno, **kwargs)



# pylint:disable=no-self-use,unused-argument,too-many-boolean-expressions
from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import MutableMapping
import logging

import archinfo

from angr.ailment.block import Block
from angr.ailment.manager import Manager
from angr.ailment.statement import (
    Statement,
    Assignment,
    CAS,
    Store,
    SideEffectStatement,
    Return,
    ConditionalJump,
    DirtyStatement,
    Jump,
    WeakAssignment,
)
from angr.ailment.expression import (
    Atom,
    Call,
    Expression,
    Extract,
    Insert,
    Register,
    VirtualVariable,
    Load,
    Const,
    VirtualVariableCategory,
    BinaryOp,
    UnaryOp,
    Convert,
    StackBaseOffset,
    VEXCCallExpression,
    ITE,
    Tmp,
    DirtyExpression,
    Reinterpret,
)

from angr.ailment.tagged_object import TaggedObject
from angr.engines.light.engine import SimEngineNostmtAIL
from .rewriting_state import RewritingState
from .consts import MAX_STACK_VAR_SIZE

if TYPE_CHECKING:
    from angr.analyses.decompiler.ssailification.ssailification import Def, UDef


_l = logging.getLogger(__name__)


class SimEngineSSARewriting(
    SimEngineNostmtAIL[RewritingState, Expression | None, Statement | tuple[Statement, ...], None]
):
    """
    This engine rewrites every block to insert phi variables and replaces every used variable with their versioned
    copies at each use location.
    """

    def __init__(
        self,
        project,
        *,
        ail_manager: Manager,
        def_to_udef: MutableMapping[Def, UDef],
        incomplete_defs: set[Def],
        vvar_id_start: int = 0,
        rewrite_tmps: bool = False,
        stackvars: bool = False,
        fail_fast: bool = False,
    ):
        super().__init__(project)

        self.def_to_vvid_cache: dict[Def, int] = {}
        self.tmp_to_vvid_cache: dict[tuple[int, int | None, int], int] = {}
        self.rewrite_tmps = rewrite_tmps
        self.ail_manager = ail_manager
        self.hclb_side_exit_state: RewritingState | None = None
        self.out_block: Block | None = None
        self.def_to_udef = def_to_udef
        self.stackvars = stackvars
        self.incomplete_defs = incomplete_defs
        self._fail_fast = fail_fast

        self._current_vvar_id = vvar_id_start
        self._extra_defs: list[int] = []

    @property
    def current_vvar_id(self) -> int:
        return self._current_vvar_id

    #
    # Util functions
    #

    @staticmethod
    def _is_head_controlled_loop_jump(block, jump_stmt: ConditionalJump) -> bool:
        concrete_targets = []
        if isinstance(jump_stmt.true_target, Const):
            concrete_targets.append(jump_stmt.true_target.value)
        if isinstance(jump_stmt.false_target, Const):
            concrete_targets.append(jump_stmt.false_target.value)
        return not all(block.addr <= t < block.addr + block.original_size for t in concrete_targets)

    #
    # Handlers
    #

    def process(
        self, state: RewritingState, *, block: Block | None = None, whitelist: set[int] | None = None, **kwargs
    ) -> None:
        super().process(state, block=block, whitelist=whitelist, **kwargs)

    def _top(self, bits):
        assert False, "Unreachable"

    def _is_top(self, expr):
        assert False, "Unreachable"

    def _process_block_end(self, block, stmt_data, whitelist):
        assert whitelist is None
        if all(stmt is None for stmt in stmt_data):
            self.out_block = block
            return

        self.out_block = Block(block.addr, block.original_size, idx=block.idx)
        for stmt_idx, new_stmt in enumerate(stmt_data):
            if new_stmt is not None:
                if isinstance(new_stmt, tuple):
                    for stmt_ in new_stmt:
                        self.out_block.statements.append(stmt_)
                else:
                    self.out_block.statements.append(new_stmt)
            else:
                self.out_block.statements.append(block.statements[stmt_idx])

    def _stmt(self, stmt: Statement):
        self._extra_defs = []
        result = super()._stmt(stmt)
        for rstmt in result if isinstance(result, tuple) else [result] if isinstance(result, Statement) else []:
            if self._extra_defs:
                rstmt.tags["extra_defs"] = self._extra_defs
            else:
                rstmt.tags.pop("extra_defs", None)

        return result

    def _handle_expr_VirtualVariable(self, expr):
        return None

    def _handle_stmt_Assignment(self, stmt):
        new_src = self._expr(stmt.src)
        new_dst = self._replace_def_expr(stmt.dst, new_src or stmt.src, stmt)
        if new_dst is not None:
            return new_dst

        if new_src is not None:
            return Assignment(
                stmt.idx,
                stmt.dst,
                new_src,
                **stmt.tags,
            )
        return None

    def _handle_stmt_WeakAssignment(self, stmt) -> WeakAssignment | None:
        new_src = self._expr(stmt.src)
        new_dst = self._expr(stmt.dst)

        if new_dst is not None or new_src is not None:
            return WeakAssignment(
                stmt.idx,
                stmt.dst if new_dst is None else new_dst,  # type: ignore
                stmt.src if new_src is None else new_src,
                **stmt.tags,
            )
        return None

    def _handle_stmt_CAS(self, stmt: CAS) -> CAS | None:
        new_addr = self._expr(stmt.addr)
        new_data_lo = self._expr(stmt.data_lo)
        new_data_hi = self._expr(stmt.data_hi) if stmt.data_hi is not None else None
        new_expd_lo = self._expr(stmt.expd_lo)
        new_expd_hi = self._expr(stmt.expd_hi) if stmt.expd_hi is not None else None
        new_old_lo = self._expr(stmt.old_lo)
        new_old_hi = self._expr(stmt.old_hi) if stmt.old_hi is not None else None
        assert new_old_lo is None or isinstance(new_old_lo, Atom)
        assert new_old_hi is None or isinstance(new_old_hi, Atom)

        if (
            new_addr is not None
            or new_old_lo is not None
            or new_old_hi is not None
            or new_data_lo is not None
            or new_data_hi is not None
            or new_expd_lo is not None
            or new_expd_hi is not None
        ):
            return CAS(
                stmt.idx,
                stmt.addr if new_addr is None else new_addr,
                stmt.data_lo if new_data_lo is None else new_data_lo,
                stmt.data_hi if new_data_hi is None else new_data_hi,
                stmt.expd_lo if new_expd_lo is None else new_expd_lo,
                stmt.expd_hi if new_expd_hi is None else new_expd_hi,
                stmt.old_lo if new_old_lo is None else new_old_lo,
                stmt.old_hi if new_old_hi is None else new_old_hi,
                stmt.endness,
                **stmt.tags,
            )
        return None

    def _handle_stmt_Store(self, stmt: Store) -> Store | Assignment | tuple[Assignment, ...] | None:
        new_data = self._expr(stmt.data)
        if self.stackvars and stmt.guard is None and isinstance(stmt.addr, StackBaseOffset):
            # vvar assignment
            vvar = self._expr_to_vvar(stmt.addr, False)
            assert isinstance(stmt.addr.offset, int)
            return self._vvar_update(vvar, stmt.addr.offset - vvar.stack_offset, new_data or stmt.data, stmt)

        # fall back to Store
        new_addr = self._expr(stmt.addr)
        new_guard = self._expr(stmt.guard) if stmt.guard is not None else None

        if new_addr is not None or new_data is not None or new_guard is not None:
            return Store(
                stmt.idx,
                stmt.addr if new_addr is None else new_addr,
                stmt.data if new_data is None else new_data,
                stmt.size,
                stmt.endness,
                guard=stmt.guard if new_guard is None else new_guard,
                **stmt.tags,
            )

        return None

    def _handle_stmt_Jump(self, stmt: Jump) -> Jump | None:
        new_target = self._expr(stmt.target)
        if new_target is not None:
            return Jump(stmt.idx, new_target, stmt.target_idx, **stmt.tags)
        return None

    def _handle_stmt_ConditionalJump(self, stmt: ConditionalJump) -> ConditionalJump | None:
        new_cond = self._expr(stmt.condition)
        new_true_target = self._expr(stmt.true_target) if stmt.true_target is not None else None
        new_false_target = self._expr(stmt.false_target) if stmt.false_target is not None else None

        if self.stmt_idx != len(self.block.statements) - 1 and self._is_head_controlled_loop_jump(self.block, stmt):
            # the conditional jump is in the middle of the block (e.g., the block generated from lifting rep stosq).
            # we need to make a copy of the state and use the state of this point in its successor
            self.hclb_side_exit_state = self.state.copy()

        if new_cond is not None or new_true_target is not None or new_false_target is not None:
            return ConditionalJump(
                stmt.idx,
                stmt.condition if new_cond is None else new_cond,
                stmt.true_target if new_true_target is None else new_true_target,
                stmt.false_target if new_false_target is None else new_false_target,
                true_target_idx=stmt.true_target_idx,
                false_target_idx=stmt.false_target_idx,
                **stmt.tags,
            )
        return None

    def _handle_stmt_SideEffectStatement(self, stmt: SideEffectStatement) -> Statement:
        new_args = None
        if stmt.expr.args is not None:
            new_args = []
            for arg in stmt.expr.args:
                new_arg = self._expr(arg)
                if new_arg is not None:
                    new_args.append(new_arg)
                else:
                    new_args.append(arg)

        new_target = self._expr(stmt.expr.target) if not isinstance(stmt.expr.target, str) else None
        replaced_call = Call(
            stmt.idx,
            stmt.expr.target if new_target is None else new_target,
            calling_convention=stmt.expr.calling_convention,
            prototype=stmt.expr.prototype,
            args=new_args,
            bits=stmt.bits,
            **stmt.tags,
        )

        cc = stmt.expr.calling_convention if stmt.expr.calling_convention is not None else self.project.factory.cc()
        if cc is not None:
            # clean up all caller-saved registers (and their subregisters)
            for reg_name in cc.CALLER_SAVED_REGS:
                base_off, base_size = self.arch.registers[reg_name]
                for suboff in range(base_off, base_off + base_size):
                    self.state.registers.pop(suboff, None)

        new_stmt = None
        if stmt.ret_expr is not None:
            assert isinstance(stmt.ret_expr, Atom)
            new_stmt = self._replace_def_expr(stmt.ret_expr, replaced_call, stmt)
        elif stmt.fp_ret_expr is not None:
            assert isinstance(stmt.fp_ret_expr, Atom)
            new_stmt = self._replace_def_expr(stmt.fp_ret_expr, replaced_call, stmt)
        if new_stmt is None:
            new_stmt = SideEffectStatement(stmt.idx, replaced_call, **stmt.tags)

        return new_stmt

    def _handle_stmt_DirtyStatement(self, stmt: DirtyStatement) -> DirtyStatement | None:
        dirty = self._expr(stmt.dirty)
        if dirty is None or dirty is stmt.dirty:
            return None
        assert isinstance(dirty, DirtyExpression)
        return DirtyStatement(stmt.idx, dirty, **stmt.tags)

    def _handle_expr_Register(self, expr: Register) -> VirtualVariable | Expression | None:
        vvar = self._expr_to_vvar(expr, True)
        return self._vvar_extract(vvar, expr.size, expr.reg_offset - vvar.reg_offset, expr)

    def _handle_expr_Tmp(self, expr: Tmp) -> VirtualVariable | None:
        if not self.rewrite_tmps:
            return None
        tmp_key = self.block.addr, self.block.idx, expr.tmp_idx
        if (vvid := self.tmp_to_vvid_cache.get(tmp_key, None)) is None:
            vvid = self.tmp_to_vvid_cache[tmp_key] = self._current_vvar_id
            self._current_vvar_id += 1
        return VirtualVariable(
            expr.idx,
            vvid,
            expr.bits,
            VirtualVariableCategory.TMP,
            oident=expr.tmp_idx,
            **(expr.tags | {"ins_addr": self.ins_addr}),
        )

    def _handle_expr_Load(self, expr: Load) -> Expression | None:
        if self.stackvars and isinstance(expr.addr, StackBaseOffset):
            # vvar assignment
            vvar = self._expr_to_vvar(expr.addr, True)
            assert isinstance(expr.addr.offset, int)
            if vvar.stack_offset + vvar.size >= expr.addr.offset + expr.size:
                return self._vvar_extract(vvar, expr.size, expr.addr.offset - vvar.stack_offset, expr)

        new_addr = self._expr(expr.addr)
        if new_addr is not None:
            return Load(expr.idx, new_addr, expr.size, expr.endness, guard=expr.guard, alt=expr.alt, **expr.tags)
        return None

    def _handle_expr_Convert(self, expr: Convert) -> Convert | None:
        new_operand = self._expr(expr.operand)
        if new_operand is not None:
            return Convert(
                expr.idx,
                expr.from_bits,
                expr.to_bits,
                expr.is_signed,
                new_operand,
                from_type=expr.from_type,
                to_type=expr.to_type,
                rounding_mode=expr.rounding_mode,
                **expr.tags,
            )
        return None

    def _handle_stmt_Return(self, stmt: Return) -> Return | None:
        if stmt.ret_exprs is None:
            return None

        updated = False
        new_ret_exprs = []
        for r in stmt.ret_exprs:
            new_r = self._expr(r)
            if new_r is not None:
                updated = True
                new_ret_exprs.append(new_r)
            else:
                new_ret_exprs.append(r)

        if not updated:
            return None
        return Return(stmt.idx, new_ret_exprs, **stmt.tags)

    def _handle_expr_BinaryOp(self, expr: BinaryOp) -> BinaryOp | None:
        new_op0 = self._expr(expr.operands[0])
        new_op1 = self._expr(expr.operands[1])

        if new_op0 is not None or new_op1 is not None:
            return BinaryOp(
                expr.idx,
                expr.op,
                [
                    expr.operands[0] if new_op0 is None else new_op0,
                    expr.operands[1] if new_op1 is None else new_op1,
                ],
                expr.signed,
                bits=expr.bits,
                floating_point=expr.floating_point,
                rounding_mode=expr.rounding_mode,
                **expr.tags,
            )
        return None

    def _handle_expr_UnaryOp(self, expr) -> UnaryOp | None:
        new_op = self._expr(expr.operand)
        if new_op is not None:
            return UnaryOp(
                expr.idx,
                expr.op,
                new_op,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_ITE(self, expr: ITE) -> ITE | None:
        new_cond = self._expr(expr.cond)
        new_iftrue = self._expr(expr.iftrue)
        new_iffalse = self._expr(expr.iffalse)

        if new_cond is not None or new_iftrue is not None or new_iffalse is not None:
            return ITE(
                expr.idx,
                expr.cond if new_cond is None else new_cond,
                expr.iffalse if new_iffalse is None else new_iffalse,
                expr.iftrue if new_iftrue is None else new_iftrue,
                **expr.tags,
            )
        return None

    def _handle_expr_VEXCCallExpression(self, expr: VEXCCallExpression) -> VEXCCallExpression | None:
        updated = False
        new_operands = []
        for operand in expr.operands:
            new_operand = self._expr(operand)
            if new_operand is not None:
                updated = True
                new_operands.append(new_operand)
            else:
                new_operands.append(operand)

        if updated:
            return VEXCCallExpression(expr.idx, expr.callee, tuple(new_operands), bits=expr.bits, **expr.tags)
        return None

    def _handle_expr_BasePointerOffset(self, expr):
        return None

    def _handle_expr_Call(self, expr):
        new_args = None
        if expr.args is not None:
            new_args = []
            for arg in expr.args:
                new_arg = self._expr(arg)
                if new_arg is not None:
                    new_args.append(new_arg)
                else:
                    new_args.append(arg)

        new_target = self._expr(expr.target) if not isinstance(expr.target, str) else None
        if new_target is not None or new_args is not None:
            return Call(
                expr.idx,
                expr.target if new_target is None else new_target,
                calling_convention=expr.calling_convention,
                prototype=expr.prototype,
                args=new_args if new_args is not None else expr.args,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_Const(self, expr):
        return None

    def _handle_expr_DirtyExpression(self, expr: DirtyExpression) -> DirtyExpression | None:
        updated = False
        new_operands = []
        for operand in expr.operands:
            new_operand = self._expr(operand)
            if new_operand is not None:
                updated = True
                new_operands.append(new_operand)
            else:
                new_operands.append(operand)

        new_guard = None
        if expr.guard is not None:
            new_guard = self._expr(expr.guard)
            if new_guard is not None:
                updated = True

        if updated:
            return DirtyExpression(
                expr.idx,
                expr.callee,
                new_operands,
                guard=new_guard,
                mfx=expr.mfx,
                maddr=expr.maddr,
                msize=expr.msize,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_MultiStatementExpression(self, expr):
        return None

    def _handle_expr_Phi(self, expr):
        return None

    def _handle_expr_Reinterpret(self, expr: Reinterpret) -> Reinterpret | None:
        new_operand = self._expr(expr.operand)
        if new_operand is not None:
            return Reinterpret(
                expr.idx,
                expr.from_bits,
                expr.from_type,
                expr.to_bits,
                expr.to_type,
                new_operand,
                **expr.tags,
            )
        return None

    def _handle_expr_StackBaseOffset(self, expr):
        if not self.stackvars:
            return None

        # if we get here, it means it's NOT through a Load or Store
        # so we won't get the opportunity to lift this into an appropriate Assignment
        # so if we don't completely redefine the value we should not create a new vvar
        if expr in self.incomplete_defs and expr.offset in self.state.stackvars:
            self.def_to_udef.pop(expr, None)

        vvar = self._expr_to_vvar(expr, True)
        refers = UnaryOp(expr.idx, "Reference", vvar, bits=expr.bits, **expr.tags)
        if expr in self.def_to_udef:
            refers.tags["extra_def"] = True
            self._extra_defs.append(vvar.varid)
        if vvar.stack_offset == expr.offset:
            return refers

        return BinaryOp(expr.idx, "Add", [refers, Const(None, None, vvar.stack_offset - expr.offset, refers.bits)])

    def _handle_expr_Extract(self, expr: Extract):
        base = self._expr(expr.base) or expr.base
        offset = self._expr(expr.offset) or expr.offset

        if base is not expr.base or offset is not expr.offset:
            return Extract(expr.idx, expr.bits, base, offset, expr.endness, **expr.tags)
        return None

    def _handle_expr_Insert(self, expr: Insert):
        base = self._expr(expr.base) or expr.base
        offset = self._expr(expr.offset) or expr.offset
        value = self._expr(expr.value) or expr.value

        if base is not expr.base or offset is not expr.offset or value is not expr.value:
            return Insert(expr.idx, base, offset, value, expr.endness, **expr.tags)
        return None

    #
    # Expression replacement
    #

    def _replace_def_expr(self, thing: Atom, value: Expression, orig_tags: TaggedObject) -> Assignment | None:
        """
        Return a new virtual variable for the given defined expression.
        """
        if isinstance(thing, Register):
            return self._replace_def_reg(thing, value, orig_tags)
        if isinstance(thing, Tmp) and self.rewrite_tmps:
            return self._replace_def_tmp(thing, value, orig_tags)
        if isinstance(thing, VirtualVariable):
            # update liveness info
            if thing.category == VirtualVariableCategory.REGISTER:
                for suboff in range(thing.reg_offset, thing.reg_offset + thing.size):
                    self.state.registers[suboff] = thing
            elif thing.category == VirtualVariableCategory.STACK:
                for suboff in range(thing.stack_offset, thing.stack_offset + thing.size):
                    self.state.stackvars[suboff] = thing
        return None

    def _replace_def_reg(self, expr: Register, value: Expression, orig_tags: TaggedObject) -> Assignment:
        """
        Return a new virtual variable for the given defined register.
        """
        vvar = self._expr_to_vvar(expr, False)
        return self._vvar_update(vvar, expr.reg_offset - vvar.reg_offset, value, orig_tags)

    def _replace_def_tmp(self, expr: Tmp, value: Expression, orig_tags: TaggedObject) -> Assignment:
        if not self.rewrite_tmps:
            return Assignment(orig_tags.idx, expr, value, **orig_tags.tags)
        vvar = self._handle_expr_Tmp(expr)
        assert vvar is not None
        return self._vvar_update(vvar, 0, value, orig_tags)

    #
    # Utils
    #

    def _expr_to_vvar(self, expr: Def, def_is_implicit: bool) -> VirtualVariable:
        # is this a use, not a def?
        if (udef := self.def_to_udef.get(expr, None)) is None:
            # in case of emergency, raise keyerror
            if isinstance(expr, StackBaseOffset):
                assert isinstance(expr.offset, int)
                if self._fail_fast or expr.offset in self.state.stackvars:
                    return self.state.stackvars[expr.offset]
            elif isinstance(expr, Register):
                if self._fail_fast or expr.reg_offset in self.state.registers:
                    return self.state.registers[expr.reg_offset]
            else:
                raise TypeError(expr)

            # we got here because expr refers to a non-existent stack offset or register offset.
            # raise a KeyError if fail_fast is specified because something else has gone wrong at this point.
            if self._fail_fast:
                raise KeyError(expr)
            # otherwise, we try our best to guesstimate the udef here
            kind = "stack" if isinstance(expr, StackBaseOffset) else "reg"
            offset = expr.offset if isinstance(expr, StackBaseOffset) else expr.reg_offset
            if kind == "stack":
                next_off = min((o for o in self.state.stackvars if o >= offset), default=offset + 4)
            else:
                # kind == "reg"
                next_off = min((o for o in self.state.registers if o >= offset), default=offset + 4)
            size = next_off - offset
        else:
            # unpack udef
            kind, offset, size = udef

        if (varid := self.def_to_vvid_cache.get(expr, None)) is None:
            varid = self.def_to_vvid_cache[expr] = self._current_vvar_id
            self._current_vvar_id += 1
        idx = self.ail_manager.next_atom()
        # TODO replace these str kinds with VirtualVariableCategory I guess
        if kind == "stack":
            category = VirtualVariableCategory.STACK
            oident = offset
        elif kind == "reg":
            category = VirtualVariableCategory.REGISTER
            oident = offset
        else:
            raise TypeError(expr)

        if kind == "stack" and size >= MAX_STACK_VAR_SIZE:
            # limit the stack variable size
            size = MAX_STACK_VAR_SIZE

        vvar = VirtualVariable(idx, varid, size * 8, category, oident, **(expr.tags | {"ins_addr": self.ins_addr}))
        if def_is_implicit:
            if kind == "stack":
                for suboff in range(offset, offset + size):
                    self.state.stackvars[suboff] = vvar
            elif kind == "reg":
                for suboff in range(offset, offset + size):
                    self.state.registers[suboff] = vvar
        return vvar

    def _vvar_extract(
        self, vvar: VirtualVariable, size: int, offset: int, orig_tags: TaggedObject
    ) -> Extract | VirtualVariable | BinaryOp:
        assert offset >= 0
        if size == vvar.size:
            return vvar
        endness = (
            self.project.arch.memory_endness
            if vvar.was_stack or (vvar.was_parameter and vvar.parameter_category == VirtualVariableCategory.STACK)
            else self.project.arch.register_endness
        )
        if size > vvar.size:
            if self._fail_fast:
                assert False, "Invariant failure: we generated a vvar which is smaller than one of its uses"
            remainder = Const(None, None, 0, size * 8 - vvar.bits, uninitalized=True)
            order = [vvar, remainder] if endness == archinfo.Endness.LE else [remainder, vvar]
            return BinaryOp(
                self.ail_manager.next_atom(),
                "Concat",
                order,
                bits=size * 8,
            )
        return Extract(
            self.ail_manager.next_atom(), size * 8, vvar, Const(None, None, offset, 64), endness, **orig_tags.tags
        )

    def _vvar_update(
        self, vvar: VirtualVariable, offset: int, value: Expression, orig_tags: TaggedObject
    ) -> Assignment:
        assert offset >= 0
        if value.bits == vvar.bits:
            combined = value
        else:
            if vvar.category == VirtualVariableCategory.STACK:
                base = self.state.stackvars.get(vvar.stack_offset, None)
            elif vvar.category == VirtualVariableCategory.REGISTER:
                base = self.state.registers.get(vvar.reg_offset, None)
            else:
                raise TypeError(vvar.category)
            if base is None:
                base = Const(None, None, 0, vvar.bits, uninitialized=True)
            endness = (
                self.project.arch.memory_endness
                if vvar.was_stack or (vvar.was_parameter and vvar.parameter_category == VirtualVariableCategory.STACK)
                else self.project.arch.register_endness
            )
            if base.bits < vvar.bits:
                base = BinaryOp(
                    self.ail_manager.next_atom(),
                    "Concat",
                    [base, Const(None, None, 0, vvar.bits - base.bits, uninitialized=True)],
                    bits=vvar.bits,
                )
            elif base.bits > vvar.bits:
                base = Extract(self.ail_manager.next_atom(), vvar.bits, base, Const(None, None, offset, 64), endness)
            combined = Insert(self.ail_manager.next_atom(), base, Const(None, None, offset, 64), value, endness)

        if vvar.category == VirtualVariableCategory.STACK:
            for suboff in range(vvar.stack_offset, vvar.stack_offset + vvar.size):
                self.state.stackvars[suboff] = vvar
        elif vvar.category == VirtualVariableCategory.REGISTER:
            for suboff in range(vvar.reg_offset, vvar.reg_offset + vvar.size):
                self.state.registers[suboff] = vvar
        return Assignment(self.ail_manager.next_atom(), vvar, combined, **orig_tags.tags)

    def _unreachable(self, *args, **kwargs):
        assert False

    _handle_binop_Add = _unreachable
    _handle_binop_AddF = _unreachable
    _handle_binop_AddV = _unreachable
    _handle_binop_And = _unreachable
    _handle_binop_Carry = _unreachable
    _handle_binop_CmpEQ = _unreachable
    _handle_binop_CmpF = _unreachable
    _handle_binop_CmpGE = _unreachable
    _handle_binop_CmpGT = _unreachable
    _handle_binop_CmpLE = _unreachable
    _handle_binop_CmpLT = _unreachable
    _handle_binop_CmpNE = _unreachable
    _handle_binop_CmpORD = _unreachable
    _handle_binop_Concat = _unreachable
    _handle_binop_Div = _unreachable
    _handle_binop_DivF = _unreachable
    _handle_binop_DivV = _unreachable
    _handle_binop_LogicalAnd = _unreachable
    _handle_binop_LogicalOr = _unreachable
    _handle_binop_Mod = _unreachable
    _handle_binop_Mul = _unreachable
    _handle_binop_Mull = _unreachable
    _handle_binop_MulF = _unreachable
    _handle_binop_MulV = _unreachable
    _handle_binop_MulHiV = _unreachable
    _handle_binop_Or = _unreachable
    _handle_binop_Rol = _unreachable
    _handle_binop_Ror = _unreachable
    _handle_binop_SBorrow = _unreachable
    _handle_binop_SCarry = _unreachable
    _handle_binop_Sar = _unreachable
    _handle_binop_Shl = _unreachable
    _handle_binop_Shr = _unreachable
    _handle_binop_Sub = _unreachable
    _handle_binop_SubF = _unreachable
    _handle_binop_SubV = _unreachable
    _handle_binop_Xor = _unreachable
    _handle_binop_InterleaveLOV = _unreachable
    _handle_binop_InterleaveHIV = _unreachable
    _handle_binop_CasCmpEQ = _unreachable
    _handle_binop_CasCmpNE = _unreachable
    _handle_binop_ExpCmpNE = _unreachable
    _handle_binop_SarNV = _unreachable
    _handle_binop_ShrNV = _unreachable
    _handle_binop_ShlNV = _unreachable
    _handle_binop_CmpEQV = _unreachable
    _handle_binop_CmpNEV = _unreachable
    _handle_binop_CmpGEV = _unreachable
    _handle_binop_CmpGTV = _unreachable
    _handle_binop_CmpLEV = _unreachable
    _handle_binop_CmpLTV = _unreachable
    _handle_binop_MinV = _unreachable
    _handle_binop_MaxV = _unreachable
    _handle_binop_QAddV = _unreachable
    _handle_binop_QNarrowBinV = _unreachable
    _handle_binop_PermV = _unreachable
    _handle_binop_Set = _unreachable
    _handle_unop_BitwiseNeg = _unreachable
    _handle_unop_Dereference = _unreachable
    _handle_unop_Neg = _unreachable
    _handle_unop_Not = _unreachable
    _handle_unop_Reference = _unreachable
    _handle_unop_Clz = _unreachable
    _handle_unop_Ctz = _unreachable
    _handle_unop_GetMSBs = _unreachable
    _handle_unop_unpack = _unreachable
    _handle_unop_Sqrt = _unreachable
    _handle_unop_RSqrtEst = _unreachable



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Semantic variable naming for size/length/count variables.

This module detects variables used as sizes, lengths, or counts and names them
appropriately (e.g., size, len, count, n).
"""

from __future__ import annotations
import logging
from collections import defaultdict

from angr.ailment import Block
from angr.ailment.statement import SideEffectStatement
from angr.sim_variable import SimVariable

from .naming_base import ClinicNamingBase

l = logging.getLogger(name=__name__)


# Functions that take size/length parameters (param_index: name)
SIZE_PARAM_FUNCTIONS = {
    # Memory functions (param positions are 0-indexed)
    "malloc": {0: "size"},
    "calloc": {0: "count", 1: "size"},
    "realloc": {1: "size"},
    "memcpy": {2: "n"},
    "memmove": {2: "n"},
    "memset": {2: "n"},
    "memcmp": {2: "n"},
    # String functions
    "strncpy": {2: "n"},
    "strncat": {2: "n"},
    "strncmp": {2: "n"},
    "strnlen": {1: "n"},
    # I/O functions
    "read": {2: "count"},
    "write": {2: "count"},
    "fread": {1: "size", 2: "count"},
    "fwrite": {1: "size", 2: "count"},
    "recv": {2: "len"},
    "send": {2: "len"},
    "recvfrom": {2: "len"},
    "sendto": {2: "len"},
    # Buffer operations
    "snprintf": {1: "size"},
    "vsnprintf": {1: "size"},
    "fgets": {1: "size"},
}


class SizeNaming(ClinicNamingBase):
    """
    Detects variables used as sizes, lengths, or counts based on:
    1. Being passed as size parameters to known functions
    2. Being used in loop bounds (compared with loop counters)
    3. Having names that suggest size semantics
    """

    PRIORITY = 70  # After call results

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._size_vars: dict[SimVariable, str] = {}  # var -> suggested name
        self._var_usage_count: dict[SimVariable, int] = defaultdict(int)

    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze the graph for size/length variable patterns.

        :return: Dictionary mapping SimVariable to new name
        """
        if not self._graph:
            return {}

        # Find variables passed as size parameters to known functions
        self._find_size_params()

        # Convert to var_to_new_name
        self._var_to_new_name = dict(self._size_vars)

        return self._var_to_new_name

    def _find_size_params(self) -> None:
        """
        Find variables passed as size parameters to known functions.
        """
        for node in self._graph:
            if not isinstance(node, Block):
                continue

            for stmt in node.statements:
                if isinstance(stmt, SideEffectStatement):
                    self._analyze_call_params(stmt.expr)

    def _analyze_call_params(self, call: SideEffectStatement) -> None:
        """
        Analyze a function call for size parameters.
        """
        func_name = self._get_function_name(call)
        if func_name is None:
            return
        normalized = self._normalize_name(func_name)

        # Check if we have size parameter info for this function
        if normalized not in SIZE_PARAM_FUNCTIONS:
            # Check for partial matches
            for pattern in SIZE_PARAM_FUNCTIONS:
                if pattern in normalized:
                    normalized = pattern
                    break
            else:
                return

        param_info = SIZE_PARAM_FUNCTIONS[normalized]

        # Check each argument
        if call.args is None:
            return

        for param_idx, suggested_name in param_info.items():
            if param_idx < len(call.args):
                arg = call.args[param_idx]
                var = self._get_linked_variable(arg)
                if var is not None:
                    self._record_size_var(var, suggested_name)

    def _record_size_var(self, var: SimVariable, suggested_name: str) -> None:
        """
        Record a variable as a size/length variable.
        """
        if var not in self._size_vars:
            self._size_vars[var] = suggested_name
            self._var_usage_count[var] += 1
        else:
            # If already recorded, increment usage count
            self._var_usage_count[var] += 1



from __future__ import annotations
from collections.abc import Callable, Sequence
from collections.abc import Iterator

import claripy

import angr
from angr import sim_type
from angr.analyses import Analysis
from angr import ailment
from angr.ailment.block_walker import AILBlockViewer
from angr.errors import AngrCallableError
from angr.knowledge_plugins.functions.function import Function


class HashLookupAPIDeobfuscator(Analysis):
    """
    An analysis that finds functions accessing loader metadata which take concrete arguments and executes them to
    see if they resolve symbols.
    """

    def __init__(
        self, lifter: Callable[[Function], angr.analyses.decompiler.Clinic], func_addrs: Sequence[int] | None = None
    ):
        self.lifter = lifter
        self.results: dict[int, tuple[str, str]] = {}

        candidates_l0 = set()
        func_addrs = list(func_addrs or sorted(self.kb.functions))
        for idx, func_addr in enumerate(func_addrs):
            self._update_progress(0.0 + 20.0 * idx / len(func_addrs), "Finding l0 candidates")
            func = self.kb.functions.get_by_addr(func_addr)
            if self._is_metadata_accessor_candidate(func):
                candidates_l0.add(func_addr)

        # Consider predecessors to handle metadata loader wrappers
        # TODO: Constrain this more efficiently
        candidates_l1 = {p for c in candidates_l0 for p in self.kb.functions.callgraph.predecessors(c)}
        candidates_exec = []
        working = sorted(candidates_l0 | candidates_l1)
        for idx, func_addr in enumerate(working):
            self._update_progress(
                20.0 + 20.0 * idx / len(working), f"Finding execution candidates [{idx + 1}/{len(working)}]"
            )
            candidates_exec.extend(self._analyze1(self.kb.functions.get_by_addr(func_addr)))

        for idx, (a, b, c) in enumerate(candidates_exec):
            self._update_progress(
                40.0 + 60.0 * idx / len(candidates_exec), f"Executing candidates [{idx + 1}/{len(candidates_exec)}]"
            )
            self._analyze2(a, b, c)

        self._finish_progress()
        self.kb.obfuscations.type3_deobfuscated_apis.update(self.results)

    def _is_metadata_accessor_candidate(self, function: Function) -> bool:
        if function.is_simprocedure or function.is_plt or function.is_alignment:
            return False
        clinic = self.lifter(function)
        assert clinic.graph is not None
        walker0 = FindCallsTo(target="NtGetCurrentPeb")
        for node in clinic.graph:
            walker0.walk(node)
        return bool(walker0.found_calls)

    def _analyze1(self, function: Function) -> Iterator[tuple[int, Callable[..., claripy.ast.BV], list[int]]]:
        clinic = self.lifter(function)
        assert clinic.graph is not None

        callgraph = self.kb.functions.callgraph
        if function.addr not in callgraph:
            return
        walker1 = FindCallsTo(target=function.addr)
        seen = set()
        callers = sorted(set(callgraph.predecessors(function.addr)))
        for caller_addr in callers:
            if caller_addr in seen:
                continue
            seen.add(caller_addr)
            pred_func = self.kb.functions[caller_addr]
            pred_clinic = self.lifter(pred_func)
            assert pred_clinic.graph is not None
            for each_pred_node in pred_clinic.graph:
                walker1.walk(each_pred_node)

        callme = self.project.factory.callable(
            function.addr,
            prototype=function.prototype,
            cc=function.calling_convention,
            add_options=angr.options.unicorn
            | {angr.options.ZERO_FILL_UNCONSTRAINED_MEMORY, angr.options.STRICT_PAGE_ACCESS},
        )
        for _, _, call_expr in walker1.found_calls:
            conc_args = []
            for arg in call_expr.args or []:
                if not isinstance(arg, ailment.expression.Const):
                    break
                conc_args.append(arg.value)
            else:
                yield call_expr.tags["ins_addr"], callme, conc_args  # type: ignore

    def _analyze2(self, ins_addr: int, callme: Callable[..., claripy.ast.BV], conc_args: list[int]):
        try:
            result_bv: claripy.ast.BV = callme(*conc_args)  # type: ignore
        except AngrCallableError:
            return
        if result_bv.concrete and (result_sym := self.project.loader.find_symbol(result_bv.concrete_value)) is not None:
            self.results[ins_addr] = (result_sym.owner.provides or "", result_sym.name)


class FindCallsTo(AILBlockViewer):
    """
    Walker which stores calls with a given target.
    """

    def __init__(self, *args, target: str | int, **kwargs):
        super().__init__(*args, **kwargs)
        self.found_calls: list[tuple[ailment.Block, int, ailment.expression.Call]] = []
        self.target = target

    def _handle_SideEffectStatement(
        self, stmt_idx: int, stmt: ailment.statement.SideEffectStatement, block: ailment.Block | None
    ):
        # if I try to make this more readable, pre-commit changes it back to this nonsense...
        # pylint: disable=too-many-boolean-expressions
        if (
            (isinstance(self.target, str) and stmt.expr.target == self.target)
            or (
                isinstance(self.target, int)
                and isinstance(stmt.expr.target, ailment.expression.Const)
                and stmt.expr.target.value == self.target
            )
            or (
                isinstance(self.target, sim_type.SimType)
                and stmt.expr.prototype is not None
                and stmt.expr.prototype.returnty == self.target
            )
        ):
            assert block is not None
            self.found_calls.append((block, stmt_idx, stmt.expr))

        return super()._handle_SideEffectStatement(stmt_idx, stmt, block)

    def _handle_CallExpr(
        self,
        expr_idx: int,
        expr: ailment.expression.Call,
        stmt_idx: int,
        stmt: ailment.Statement | None,
        block: ailment.Block | None,
    ):
        if (isinstance(self.target, str) and expr.target == self.target) or (
            isinstance(self.target, int)
            and isinstance(expr.target, ailment.expression.Const)
            and expr.target.value == self.target
        ):
            assert block is not None
            self.found_calls.append((block, stmt_idx, expr))
        return super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)



# pylint:disable=unused-argument,no-self-use,too-many-boolean-expressions
from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from angr.ailment.block import Block
from angr.ailment.statement import (
    Statement,
    Assignment,
    Store,
    SideEffectStatement,
    CAS,
    Return,
    ConditionalJump,
    DirtyStatement,
    WeakAssignment,
)
from angr.ailment.expression import (
    Atom,
    Call,
    Expression,
    Extract,
    Insert,
    VirtualVariable,
    Load,
    BinaryOp,
    UnaryOp,
    Phi,
    Convert,
    ITE,
    VEXCCallExpression,
    DirtyExpression,
    Reinterpret,
)
from angr.engines.light import SimEngineNostmtAIL

if TYPE_CHECKING:
    from angr import KnowledgeBase


_l = logging.getLogger(__name__)


class SimEngineDephiRewriting(SimEngineNostmtAIL[None, Expression | None, Statement | tuple[Statement, ...], None]):
    """
    This engine rewrites every block to insert phi variables and replaces every used variable with their versioned
    copies at each use location.
    """

    state: None

    def __init__(
        self,
        project,
        vvar_to_vvar: dict[int, int],
        func_addr: int | None = None,
        variable_kb: KnowledgeBase | None = None,
    ):
        super().__init__(project)

        self.vvar_to_vvar = vvar_to_vvar
        self.out_block = None
        self.func_addr = func_addr
        self.variable_kb = variable_kb

        self._stmt_handlers["IncompleteSwitchCaseHeadStatement"] = self._handle_stmt_IncompleteSwitchCaseHeadStatement

    def _top(self, bits):
        assert False, "Unreachable"

    def _is_top(self, expr):
        return False

    def append_statement(self, stmt: Statement) -> None:
        if self.out_block is None:
            self.out_block = Block(self.block.addr, self.block.original_size, statements=[], idx=self.block.idx)
        self.out_block.statements.append(stmt)

    #
    # Handlers
    #

    def _process_block_end(self, block, stmt_data, whitelist):
        assert whitelist is None
        for stmt_idx, new_stmt in enumerate(stmt_data):
            if new_stmt is not None:
                if isinstance(new_stmt, tuple):
                    for stmt_ in new_stmt:
                        self.append_statement(stmt_)
                else:
                    self.append_statement(new_stmt)
            else:
                self.append_statement(block.statements[stmt_idx])

    def _handle_stmt_Assignment(self, stmt):
        new_src = self._expr(stmt.src)
        new_dst = None

        if isinstance(stmt.dst, VirtualVariable) and stmt.dst.varid in self.vvar_to_vvar:
            new_varid = self.vvar_to_vvar[stmt.dst.varid]
            new_dst = VirtualVariable(
                stmt.dst.idx,
                new_varid,
                stmt.dst.bits,
                stmt.dst.category,
                oident=stmt.dst.oident,
                variable=stmt.dst.variable,
                variable_offset=stmt.dst.variable_offset,
                **stmt.dst.tags,
            )

        # ensure we do not generate vvar_A = vvar_A or var_A = var_A (even if lhs and rhs are different vvars, they
        # can be mapped to the same variable)
        dst = stmt.dst if new_dst is None else new_dst
        src = stmt.src if new_src is None else new_src
        if isinstance(dst, VirtualVariable) and isinstance(src, VirtualVariable):
            if dst.varid == src.varid:
                # skip it
                return ()
            if (
                self.func_addr is not None
                and self.variable_kb is not None
                and self.func_addr in self.variable_kb.variables
            ):
                dst_var = getattr(dst, "variable", None)
                src_var = getattr(src, "variable", None)
                var_manager = self.variable_kb.variables[self.func_addr]
                if (
                    dst_var is not None
                    and src_var is not None
                    and var_manager.unified_variable(dst_var) is var_manager.unified_variable(src_var)
                ):
                    # skip it
                    return ()

            return Assignment(stmt.idx, dst, src, **stmt.tags)
        return None

    def _handle_stmt_WeakAssignment(self, stmt) -> WeakAssignment | None:
        new_src = self._expr(stmt.src)
        new_dst = self._expr(stmt.dst)

        if new_dst is not None or new_src is not None:
            return WeakAssignment(
                stmt.idx,
                stmt.dst if new_dst is None else new_dst,  # type: ignore
                stmt.src if new_src is None else new_src,
                **stmt.tags,
            )
        return None

    def _handle_stmt_CAS(self, stmt: CAS) -> CAS | None:
        new_addr = self._expr(stmt.addr)
        new_data_lo = self._expr(stmt.data_lo)
        new_data_hi = self._expr(stmt.data_hi) if stmt.data_hi is not None else None
        new_expd_lo = self._expr(stmt.expd_lo)
        new_expd_hi = self._expr(stmt.expd_hi) if stmt.expd_hi is not None else None
        new_old_lo = self._expr(stmt.old_lo)
        new_old_hi = self._expr(stmt.old_hi) if stmt.old_hi is not None else None
        assert new_old_lo is None or isinstance(new_old_lo, Atom)
        assert new_old_hi is None or isinstance(new_old_hi, Atom)

        if (
            new_addr is not None
            or new_old_lo is not None
            or new_old_hi is not None
            or new_data_lo is not None
            or new_data_hi is not None
            or new_expd_lo is not None
            or new_expd_hi is not None
        ):
            return CAS(
                stmt.idx,
                stmt.addr if new_addr is None else new_addr,
                stmt.data_lo if new_data_lo is None else new_data_lo,
                stmt.data_hi if new_data_hi is None else new_data_hi,
                stmt.expd_lo if new_expd_lo is None else new_expd_lo,
                stmt.expd_hi if new_expd_hi is None else new_expd_hi,
                stmt.old_lo if new_old_lo is None else new_old_lo,
                stmt.old_hi if new_old_hi is None else new_old_hi,
                stmt.endness,
                **stmt.tags,
            )
        return None

    def _handle_stmt_Store(self, stmt):
        new_addr = self._expr(stmt.addr)
        new_data = self._expr(stmt.data)

        if new_addr is not None or new_data is not None:
            return Store(
                stmt.idx,
                stmt.addr if new_addr is None else new_addr,
                stmt.data if new_data is None else new_data,
                stmt.size,
                stmt.endness,
                variable=stmt.variable,
                guard=stmt.guard,
                **stmt.tags,
            )

        return None

    def _handle_stmt_ConditionalJump(self, stmt):
        new_cond = self._expr(stmt.condition)
        new_true_target = self._expr(stmt.true_target) if stmt.true_target is not None else None
        new_false_target = self._expr(stmt.false_target) if stmt.false_target is not None else None

        if new_cond is not None or new_true_target is not None or new_false_target is not None:
            return ConditionalJump(
                stmt.idx,
                stmt.condition if new_cond is None else new_cond,
                stmt.true_target if new_true_target is None else new_true_target,
                stmt.false_target if new_false_target is None else new_false_target,
                true_target_idx=stmt.true_target_idx,
                false_target_idx=stmt.false_target_idx,
                **stmt.tags,
            )
        return None

    def _handle_stmt_SideEffectStatement(self, stmt: SideEffectStatement):
        new_target = (
            self._expr(stmt.expr.target)
            if stmt.expr.target is not None and not isinstance(stmt.expr.target, str)
            else None
        )
        new_ret_expr = self._expr(stmt.ret_expr) if stmt.ret_expr is not None else None
        new_fp_ret_expr = self._expr(stmt.fp_ret_expr) if stmt.fp_ret_expr is not None else None

        if new_target is not None or new_ret_expr is not None or new_fp_ret_expr is not None:
            return SideEffectStatement(
                stmt.idx,
                Call(
                    stmt.idx,
                    stmt.expr.target if new_target is None else new_target,
                    calling_convention=stmt.expr.calling_convention,
                    prototype=stmt.expr.prototype,
                    args=stmt.expr.args,
                    bits=stmt.bits,
                    **stmt.tags,
                ),
                ret_expr=stmt.ret_expr if new_ret_expr is None else new_ret_expr,
                fp_ret_expr=stmt.fp_ret_expr if new_fp_ret_expr is None else new_fp_ret_expr,
                **stmt.tags,
            )
        return None

    def _handle_stmt_DirtyStatement(self, stmt: DirtyStatement) -> DirtyStatement | None:
        dirty = self._expr(stmt.dirty)
        if dirty is None or dirty is stmt.dirty:
            return None
        assert isinstance(dirty, DirtyExpression)
        return DirtyStatement(stmt.idx, dirty, **stmt.tags)

    def _handle_expr_Load(self, expr):
        new_addr = self._expr(expr.addr)
        if new_addr is not None:
            return Load(expr.idx, new_addr, expr.size, expr.endness, guard=expr.guard, alt=expr.alt, **expr.tags)
        return None

    def _handle_expr_Convert(self, expr: Convert) -> Convert | None:
        new_operand = self._expr(expr.operand)
        if new_operand is not None:
            return Convert(
                expr.idx,
                expr.from_bits,
                expr.to_bits,
                expr.is_signed,
                new_operand,
                from_type=expr.from_type,
                to_type=expr.to_type,
                rounding_mode=expr.rounding_mode,
                **expr.tags,
            )
        return None

    def _handle_expr_Reinterpret(self, expr: Reinterpret) -> Reinterpret | None:
        new_operand = self._expr(expr.operand)
        if new_operand is not None:
            return Reinterpret(
                expr.idx,
                expr.from_bits,
                expr.from_type,
                expr.to_bits,
                expr.to_type,
                new_operand,
                **expr.tags,
            )
        return None

    def _handle_expr_Const(self, expr):
        return None

    def _handle_expr_Phi(self, expr: Phi) -> None:
        return None

    def _handle_expr_VirtualVariable(self, expr: VirtualVariable) -> VirtualVariable | None:
        if expr.varid in self.vvar_to_vvar:
            return VirtualVariable(
                expr.idx,
                self.vvar_to_vvar[expr.varid],
                expr.bits,
                expr.category,
                oident=expr.oident,
                variable=expr.variable,
                variable_offset=expr.variable_offset,
                **expr.tags,
            )
        return None

    def _handle_stmt_Return(self, stmt):
        if stmt.ret_exprs is None:
            new_ret_exprs = None
        else:
            updated = False
            new_ret_exprs = []
            for r in stmt.ret_exprs:
                new_r = self._expr(r)
                if new_r is not None:
                    updated = True
                new_ret_exprs.append(new_r if new_r is not None else None)
            if not updated:
                new_ret_exprs = None

        if new_ret_exprs:
            return Return(stmt.idx, new_ret_exprs, **stmt.tags)
        return None

    def _handle_stmt_IncompleteSwitchCaseHeadStatement(self, stmt):
        return None

    def _handle_expr_BinaryOp(self, expr):
        new_op0 = self._expr(expr.operands[0])
        new_op1 = self._expr(expr.operands[1])

        if new_op0 is not None or new_op1 is not None:
            return BinaryOp(
                expr.idx,
                expr.op,
                [
                    expr.operands[0] if new_op0 is None else new_op0,
                    expr.operands[1] if new_op1 is None else new_op1,
                ],
                expr.signed,
                bits=expr.bits,
                floating_point=expr.floating_point,
                rounding_mode=expr.rounding_mode,
                **expr.tags,
            )
        return None

    def _handle_expr_UnaryOp(self, expr):
        new_op0 = self._expr(expr.operands[0])

        if new_op0 is not None:
            return UnaryOp(
                expr.idx,
                expr.op,
                expr.operands[0] if new_op0 is None else new_op0,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_ITE(self, expr):
        new_cond = self._expr(expr.cond)
        new_iftrue = self._expr(expr.iftrue)
        new_iffalse = self._expr(expr.iffalse)

        if new_cond is not None or new_iftrue is not None or new_iffalse is not None:
            return ITE(
                expr.idx,
                expr.cond if new_cond is None else new_cond,
                expr.iftrue if new_iftrue is None else new_iftrue,
                expr.iffalse if new_iffalse is None else new_iffalse,
                **expr.tags,
            )
        return None

    def _handle_expr_Extract(self, expr):
        base = self._expr(expr.base) or expr.base
        offset = self._expr(expr.offset) or expr.offset

        if base is not expr.base or offset is not expr.offset:
            return Extract(expr.idx, expr.bits, base, offset, expr.endness, **expr.tags)
        return None

    def _handle_expr_Insert(self, expr):
        base = self._expr(expr.base) or expr.base
        offset = self._expr(expr.offset) or expr.offset
        value = self._expr(expr.value) or expr.value

        if base is not expr.base or offset is not expr.offset or value is not expr.value:
            return Insert(expr.idx, base, offset, value, expr.endness, **expr.tags)
        return None

    def _handle_VEXCCallExpression(self, expr: VEXCCallExpression) -> VEXCCallExpression | None:
        new_operands = []
        updated = False
        for o in expr.operands:
            new_o = self._expr(o)
            if new_o is not None:
                updated = True
                new_operands.append(new_o)
            else:
                new_operands.append(o)

        if updated:
            return VEXCCallExpression(
                expr.idx,
                expr.callee,
                tuple(new_operands),
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_DirtyExpression(self, expr: DirtyExpression) -> DirtyExpression | None:
        new_operands = []
        updated = False
        for o in expr.operands:
            new_o = self._expr(o)
            if new_o is not None:
                updated = True
                new_operands.append(new_o)
            else:
                new_operands.append(o)

        new_guard = None
        if expr.guard is not None:
            new_guard = self._expr(expr.guard)
            if new_guard is not None:
                updated = True

        if updated:
            return DirtyExpression(
                expr.idx,
                expr.callee,
                new_operands,
                guard=new_guard,
                mfx=expr.mfx,
                maddr=expr.maddr,
                msize=expr.msize,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_BasePointerOffset(self, expr):
        return None

    def _handle_expr_StackBaseOffset(self, expr):
        return None

    def _handle_expr_Call(self, expr: Call):
        new_target = self._expr(expr.target) if expr.target is not None and not isinstance(expr.target, str) else None

        if new_target is not None:
            return Call(
                expr.idx,
                new_target,
                calling_convention=expr.calling_convention,
                prototype=expr.prototype,
                args=expr.args,
                bits=expr.bits,
                **expr.tags,
            )
        return None

    def _handle_expr_MultiStatementExpression(self, expr):
        return None

    def _handle_expr_Register(self, expr):
        return None

    def _handle_expr_Tmp(self, expr):
        return None

    def _handle_expr_VEXCCallExpression(self, expr):
        return None

    def _unreachable(self, *args, **kwargs):
        assert False

    _handle_binop_Add = _unreachable
    _handle_binop_AddF = _unreachable
    _handle_binop_AddV = _unreachable
    _handle_binop_And = _unreachable
    _handle_binop_Carry = _unreachable
    _handle_binop_CmpEQ = _unreachable
    _handle_binop_CmpF = _unreachable
    _handle_binop_CmpGE = _unreachable
    _handle_binop_CmpGT = _unreachable
    _handle_binop_CmpLE = _unreachable
    _handle_binop_CmpLT = _unreachable
    _handle_binop_CmpNE = _unreachable
    _handle_binop_CmpORD = _unreachable
    _handle_binop_Concat = _unreachable
    _handle_binop_Div = _unreachable
    _handle_binop_DivF = _unreachable
    _handle_binop_DivV = _unreachable
    _handle_binop_LogicalAnd = _unreachable
    _handle_binop_LogicalOr = _unreachable
    _handle_binop_Mod = _unreachable
    _handle_binop_Mul = _unreachable
    _handle_binop_Mull = _unreachable
    _handle_binop_MulF = _unreachable
    _handle_binop_MulV = _unreachable
    _handle_binop_MulHiV = _unreachable
    _handle_binop_Or = _unreachable
    _handle_binop_Rol = _unreachable
    _handle_binop_Ror = _unreachable
    _handle_binop_SBorrow = _unreachable
    _handle_binop_SCarry = _unreachable
    _handle_binop_Sar = _unreachable
    _handle_binop_Shl = _unreachable
    _handle_binop_Shr = _unreachable
    _handle_binop_Sub = _unreachable
    _handle_binop_SubF = _unreachable
    _handle_binop_SubV = _unreachable
    _handle_binop_Xor = _unreachable
    _handle_binop_InterleaveLOV = _unreachable
    _handle_binop_InterleaveHIV = _unreachable
    _handle_binop_CasCmpEQ = _unreachable
    _handle_binop_CasCmpNE = _unreachable
    _handle_binop_ExpCmpNE = _unreachable
    _handle_binop_SarNV = _unreachable
    _handle_binop_ShrNV = _unreachable
    _handle_binop_ShlNV = _unreachable
    _handle_binop_CmpEQV = _unreachable
    _handle_binop_CmpNEV = _unreachable
    _handle_binop_CmpGEV = _unreachable
    _handle_binop_CmpGTV = _unreachable
    _handle_binop_CmpLEV = _unreachable
    _handle_binop_CmpLTV = _unreachable
    _handle_binop_MinV = _unreachable
    _handle_binop_MaxV = _unreachable
    _handle_binop_QAddV = _unreachable
    _handle_binop_QNarrowBinV = _unreachable
    _handle_binop_PermV = _unreachable
    _handle_binop_Set = _unreachable
    _handle_unop_BitwiseNeg = _unreachable
    _handle_unop_Dereference = _unreachable
    _handle_unop_Neg = _unreachable
    _handle_unop_Not = _unreachable
    _handle_unop_Reference = _unreachable
    _handle_unop_Clz = _unreachable
    _handle_unop_Ctz = _unreachable
    _handle_unop_GetMSBs = _unreachable
    _handle_unop_unpack = _unreachable
    _handle_unop_Sqrt = _unreachable
    _handle_unop_RSqrtEst = _unreachable



from __future__ import annotations

import logging

import claripy

import angr
from angr.storage.memory_mixins.address_concretization_mixin import MultiwriteAnnotation
from angr.sim_options import MEMORY_CHUNK_INDIVIDUAL_READS

l = logging.getLogger(name=__name__)


class memchr(angr.SimProcedure):
    # pylint:disable=arguments-differ, missing-class-docstring

    def run(self, s_addr, c_int, n):
        c = c_int[7:0]

        if self.state.solver.is_true(n == 0):
            return claripy.BVV(0, self.state.arch.bits)

        if not self.state.solver.symbolic(n):
            max_search = self.state.solver.eval(n)
        else:
            max_search = min(self.state.solver.max_int(n), self.state.libc.max_buffer_size)  # type: ignore

        if max_search == 0:
            return claripy.BVV(0, self.state.arch.bits)

        l.debug("memchr searching %d bytes for byte", max_search)

        chunk_size = None
        if MEMORY_CHUNK_INDIVIDUAL_READS in self.state.options:
            chunk_size = 1

        if self.state.solver.symbolic(n):
            l.debug("symbolic n")
            max_sym = min(self.state.solver.max_int(n), self.state.libc.max_symbolic_memchr)  # type: ignore
            a, constraints, i = self.state.memory.find(s_addr, c, max_search, max_symbolic_bytes=max_sym, default=0)
        else:
            l.debug("concrete n")
            a, constraints, i = self.state.memory.find(s_addr, c, max_search, default=0, chunk_size=chunk_size)

        if len(i) > 1:
            a = a.annotate(MultiwriteAnnotation())
            self.state.add_constraints(*constraints)

        # Ensure the found position is within the n bound
        chrpos = a - s_addr
        self.state.add_constraints(claripy.If(a != 0, claripy.ULT(chrpos, n), True))

        return a



from __future__ import annotations
import struct

from angr.ailment.expression import Reinterpret, Const

from .base import PeepholeOptimizationExprBase


class RemoveRedundantReinterprets(PeepholeOptimizationExprBase):
    """
    Simplify nested and constant Reinterpret() expressions.
    """

    __slots__ = ()

    NAME = "Simplifying nested and constant Reinterprets"
    expr_classes = (Reinterpret,)  # all expressions are allowed

    def optimize(self, expr: Reinterpret, **kwargs):
        if isinstance(expr.operand, Reinterpret):
            inner = expr.operand
            if expr.from_type == inner.to_type and expr.to_type == inner.from_type:
                return inner.operand

        if expr.from_type == "I" and expr.to_type == "F" and isinstance(expr.operand, Const):
            # replace it with a floating point constant
            if expr.operand.bits == 32:
                int_fmt = "<I"
            elif expr.operand.bits == 64:
                int_fmt = "<Q"
            else:
                raise NotImplementedError

            if expr.bits == 32:
                float_fmt = "<f"
            elif expr.bits == 64:
                float_fmt = "<d"
            else:
                raise NotImplementedError

            value = struct.unpack(float_fmt, struct.pack(int_fmt, expr.operand.value))[0]
            return Const(expr.idx, None, value, expr.bits, **expr.tags)

        return None



# pylint:disable=arguments-differ
from __future__ import annotations
from typing import TYPE_CHECKING

from angr.ailment.expression import (
    Call,
    Expression,
    BinaryOp,
    Const,
    Register,
    StackBaseOffset,
    UnaryOp,
    VirtualVariable,
)
from angr.ailment.statement import Store, Assignment, SideEffectStatement

from angr.ailment.tagged_object import TagDict
from angr.sim_type import PointerDisposition, SimTypeFunction, SimTypeLong, SimTypePointer, SimTypeWideChar
from .base import PeepholeOptimizationMultiStmtBase
from .inlined_wcscpy import InlinedWcscpy

if TYPE_CHECKING:
    from angr.ailment.statement import Statement


def match_statements(stmts: list[Statement], index: int) -> int:
    ending = index
    has_wcsncpy = False
    for i in range(index, len(stmts)):
        stmt = stmts[i]
        if isinstance(stmt, SideEffectStatement):
            if InlinedWcscpy.is_inlined_wcsncpy(stmt):
                has_wcsncpy = True
            else:
                break
        elif isinstance(stmt, Store):
            if not isinstance(stmt.data, Const):
                break
            _, off = InlinedWcscpyConsolidation._parse_addr(stmt.addr)
            if off is None:
                # unsupported offset - bail
                break
        elif (
            isinstance(stmt, Assignment)
            and isinstance(stmt.dst, VirtualVariable)
            and stmt.dst.was_stack
            and isinstance(stmt.src, Const)
        ):
            pass
        else:
            break
        ending = i + 1
    return ending - index if has_wcsncpy and ending - index >= 2 else 0


class InlinedWcscpyConsolidation(PeepholeOptimizationMultiStmtBase):
    """
    Consolidate multiple inlined wcscpy/wcsncpy calls.
    """

    __slots__ = ()

    NAME = "Consolidate multiple inlined wcsncpy calls"
    stmt_classes = (match_statements,)

    def optimize(  # type: ignore
        self, stmts: list[SideEffectStatement | Store | Assignment], stmt_idx: int | None = None, block=None, **kwargs
    ):  # pylint:disable=unused-argument
        reordered_stmts = self._reorder_stmts(stmts)
        if not reordered_stmts or len(reordered_stmts) <= 1:
            return None

        new_stmts = []
        optimized = False
        stop = False
        while not stop:
            new_stmts = []
            stop = True
            for i, stmt0 in enumerate(reordered_stmts):
                if i == len(reordered_stmts) - 1:
                    new_stmts.append(reordered_stmts[i])
                    break
                stmt1 = reordered_stmts[i + 1]
                opt_stmts = self._optimize_pair(stmt0, stmt1)
                if opt_stmts is None:
                    new_stmts.append(stmt0)
                else:
                    new_stmts += opt_stmts
                    # start again from the beginning
                    optimized = True
                    stop = False
                    reordered_stmts = new_stmts + reordered_stmts[i + 2 :]
                    break

        return new_stmts if optimized and new_stmts else None

    def _reorder_stmts(self, stmts: list[Call | Store | Assignment]) -> list[Call | Store] | None:
        """
        Order a list of statements based on ascending addresses of their destination buffers.
        """

        if not all(
            (
                InlinedWcscpy.is_inlined_wcsncpy(s)
                or (isinstance(s, Store) and isinstance(s.data, Const))
                or (
                    isinstance(s, Assignment)
                    and isinstance(s.dst, VirtualVariable)
                    and s.dst.was_stack
                    and isinstance(s.src, Const)
                )
            )
            for s in stmts
        ):
            return None
        offset_to_stmt = {}
        updated_offsets: set[int] = set()
        known_base = None
        for stmt in stmts:
            if isinstance(stmt, SideEffectStatement):
                assert (
                    stmt.expr.args is not None
                    and len(stmt.expr.args) == 3
                    and stmt.expr.args[0] is not None
                    and stmt.expr.args[2] is not None
                )
                base, off = self._parse_addr(stmt.expr.args[0])
                store_size = stmt.expr.args[2].value * 2 if isinstance(stmt.expr.args[2], Const) else None
            elif isinstance(stmt, Store):
                base, off = self._parse_addr(stmt.addr)
                store_size = stmt.size
            elif isinstance(stmt, Assignment):
                base, off = self._parse_addr(stmt.dst)
                store_size = stmt.dst.size
            else:
                # unexpected!
                return None
            if off is None or store_size is None:
                # bad offset or size - bail
                return None
            if known_base is None:
                known_base = base
            elif not base.likes(known_base):
                # bail
                return None
            if off in offset_to_stmt:
                # duplicate offset - bail
                return None
            assert isinstance(store_size, int)
            for i in range(store_size):
                if off + i in updated_offsets:
                    # overlapping store - bail
                    return None
                updated_offsets.add(off + i)

            offset_to_stmt[off] = stmt

        return [offset_to_stmt[k] for k in sorted(offset_to_stmt)]

    def _optimize_pair(
        self, last_stmt: SideEffectStatement | Store | Assignment, stmt: SideEffectStatement | Store | Assignment
    ) -> list[SideEffectStatement] | None:
        # convert (store, wcsncpy()) to (wcsncpy(), store) if they do not overlap
        if (
            isinstance(stmt, SideEffectStatement)
            and InlinedWcscpy.is_inlined_wcsncpy(stmt)
            and stmt.expr.args is not None
            and len(stmt.expr.args) == 3
            and isinstance(stmt.expr.args[2], Const)
            and isinstance(stmt.expr.args[2].value, int)
            and isinstance(last_stmt, (Store, Assignment))
        ):
            if isinstance(last_stmt, Store) and isinstance(last_stmt.data, Const):
                store_addr = last_stmt.addr
                store_size = last_stmt.size
            elif isinstance(last_stmt, Assignment):
                store_addr = last_stmt.dst
                store_size = last_stmt.dst.size
            else:
                return None
            # check if they overlap
            wcsncpy_addr = stmt.expr.args[0]
            wcsncpy_size = stmt.expr.args[2].value * 2
            delta = self._get_delta(store_addr, wcsncpy_addr)
            if delta is not None:
                if (0 <= delta <= store_size) or (delta < 0 and -delta <= wcsncpy_size):
                    # they overlap, do not switch
                    pass
                else:
                    last_stmt, stmt = stmt, last_stmt

        # swap two statements if they are out of order
        if InlinedWcscpy.is_inlined_wcsncpy(last_stmt) and InlinedWcscpy.is_inlined_wcsncpy(stmt):
            assert last_stmt.expr.args is not None and stmt.expr.args is not None
            delta = self._get_delta(last_stmt.expr.args[0], stmt.expr.args[0])
            if delta is not None and delta < 0:
                last_stmt, stmt = stmt, last_stmt

        if InlinedWcscpy.is_inlined_wcsncpy(last_stmt):
            assert last_stmt.expr.args is not None
            assert self.kb is not None
            s_last: bytes = self.kb.custom_strings[last_stmt.expr.args[1].value]
            addr_last = last_stmt.expr.args[0]
            new_str = None  # will be set if consolidation should happen

            if isinstance(stmt, SideEffectStatement) and InlinedWcscpy.is_inlined_wcsncpy(stmt):
                assert stmt.expr.args is not None
                # consolidating two calls
                s_curr: bytes = self.kb.custom_strings[stmt.expr.args[1].value]
                addr_curr = stmt.expr.args[0]
                # determine if the two addresses are consecutive
                delta = self._get_delta(addr_last, addr_curr)
                if delta is not None and delta == len(s_last):
                    # consolidate both calls!
                    new_str = s_last + s_curr
            elif isinstance(stmt, Store) and isinstance(stmt.data, Const) and isinstance(stmt.data.value, int):
                # consolidating a call and a store, in case the store statement is storing the suffix of a string (but
                # the suffix is too short to qualify an inlined strcpy optimization)
                addr_curr = stmt.addr
                delta = self._get_delta(addr_last, addr_curr)
                if delta is not None and delta == len(s_last):
                    if stmt.size == 2 and stmt.data.value == 0:
                        # it's probably the terminating null byte
                        r, s = True, b"\x00\x00"
                    else:
                        r, s = InlinedWcscpy.is_integer_likely_a_wide_string(
                            stmt.data.value,
                            stmt.size,
                            stmt.endness,
                            min_length=1,  # type: ignore
                        )
                    if r and s is not None:
                        new_str = s_last + s
            elif (
                isinstance(stmt, Assignment)
                and isinstance(stmt.dst, VirtualVariable)
                and isinstance(stmt.src, Const)
                and isinstance(stmt.src.value, int)
            ):
                # consolidating a call and an assignment, in case the assignment statement is storing the suffix of a
                # string (but the suffix is too short to qualify an inlined strcpy optimization)
                addr_curr = stmt.dst
                delta = self._get_delta(addr_last, addr_curr)
                if delta is not None and delta == len(s_last):
                    r, s = InlinedWcscpy.is_integer_likely_a_wide_string(
                        stmt.src.value,
                        stmt.dst.size,
                        self.project.arch.memory_endness,
                        min_length=1,  # type: ignore
                    )
                    if r and s is not None:
                        new_str = s_last + s

            if new_str is not None:
                assert self.project is not None
                wstr_type = SimTypePointer(SimTypeWideChar()).with_arch(self.project.arch)
                wstr_type_out = SimTypePointer(SimTypeWideChar(), disposition=PointerDisposition.OUT)
                prototype = SimTypeFunction([wstr_type_out, wstr_type, SimTypeLong(signed=False)], wstr_type).with_arch(
                    self.project.arch
                )
                if new_str.endswith(b"\x00\x00"):
                    call_name = "wcsncpy"
                    new_str_idx = self.kb.custom_strings.allocate(new_str[:-2])
                    args = [
                        last_stmt.expr.args[0],
                        Const(None, None, new_str_idx, last_stmt.expr.args[0].bits, custom_string=True, type=wstr_type),
                    ]
                else:
                    call_name = "wcsncpy"
                    new_str_idx = self.kb.custom_strings.allocate(new_str)
                    args = [
                        last_stmt.expr.args[0],
                        Const(None, None, new_str_idx, last_stmt.expr.args[0].bits, custom_string=True, type=wstr_type),
                        Const(None, None, len(new_str) // 2, self.project.arch.bits),
                    ]

                tags = TagDict(stmt.tags)
                if args[0].tags.get("extra_def", False):
                    assert isinstance(args[0], UnaryOp)
                    assert args[0].op == "Reference"
                    assert isinstance(args[0].operand, VirtualVariable)
                    tags["extra_defs"] = [args[0].operand.varid]
                else:
                    tags.pop("extra_defs", None)
                return [
                    SideEffectStatement(
                        stmt.idx, Call(stmt.idx, call_name, args=args, prototype=prototype, **tags), **tags
                    )
                ]

        return None

    @staticmethod
    def _parse_addr(addr: Expression) -> tuple[Expression, int]:
        # we force the base to 64-bit because it does not really matter when we use it

        if isinstance(addr, VirtualVariable) and addr.was_stack:
            return StackBaseOffset(-1, 64, 0), addr.stack_offset
        if isinstance(addr, Register):
            return addr, 0
        if isinstance(addr, StackBaseOffset):
            return StackBaseOffset(-1, 64, 0), addr.offset
        if (
            isinstance(addr, UnaryOp)
            and addr.op == "Reference"
            and isinstance(addr.operand, VirtualVariable)
            and addr.operand.was_stack
        ):
            return StackBaseOffset(-1, 64, 0), addr.operand.stack_offset
        if isinstance(addr, BinaryOp):
            if addr.op == "Add" and isinstance(addr.operands[1], Const) and isinstance(addr.operands[1].value, int):
                base_0, offset_0 = InlinedWcscpyConsolidation._parse_addr(addr.operands[0])
                return base_0, offset_0 + addr.operands[1].value
            if addr.op == "Sub" and isinstance(addr.operands[1], Const) and isinstance(addr.operands[1].value, int):
                base_0, offset_0 = InlinedWcscpyConsolidation._parse_addr(addr.operands[0])
                return base_0, offset_0 - addr.operands[1].value

        return addr, 0

    @staticmethod
    def _get_delta(addr_0: Expression, addr_1: Expression) -> int | None:
        base_0, offset_0 = InlinedWcscpyConsolidation._parse_addr(addr_0)
        base_1, offset_1 = InlinedWcscpyConsolidation._parse_addr(addr_1)
        if base_0.likes(base_1):
            return offset_1 - offset_0
        return None



# pylint:disable=missing-class-docstring
from __future__ import annotations
from enum import Enum
from itertools import count

from sortedcontainers import SortedDict

from angr.sim_variable import SimVariable

IdentType = tuple[int, int, str]


class CConstantType(Enum):
    """
    Describes the value type of a CConstant class.
    """

    INT = 0
    FLOAT = 1
    STRING = 2


#
#   Position Mapping Classes
#


class PositionMappingElement:
    __slots__ = ("length", "obj", "start")

    def __init__(self, start, length, obj):
        self.start: int = start
        self.length: int = length
        self.obj = obj

    def __contains__(self, offset):
        return self.start <= offset < self.start + self.length

    def __repr__(self):
        return f"<{self.start}-{self.start + self.length}: {self.obj}>"


class PositionMapping:
    __slots__ = ("_posmap",)

    DUPLICATION_CHECK = True

    def __init__(self):
        self._posmap: SortedDict | dict[int, PositionMappingElement] = SortedDict()

    def items(self):
        return self._posmap.items()

    #
    # Public methods
    #

    def add_mapping(self, start_pos, length, obj):
        # duplication check
        if self.DUPLICATION_CHECK:
            try:
                pre = next(self._posmap.irange(maximum=start_pos, reverse=True))
                if start_pos in self._posmap[pre]:
                    raise ValueError("New mapping is overlapping with an existing element.")
            except StopIteration:
                pass

        self._posmap[start_pos] = PositionMappingElement(start_pos, length, obj)

    def get_node(self, pos: int):
        element = self.get_element(pos)
        if element is None:
            return None
        return element.obj

    def get_element(self, pos: int) -> PositionMappingElement | None:
        try:
            pre = next(self._posmap.irange(maximum=pos, reverse=True))
        except StopIteration:
            return None

        element = self._posmap[pre]
        if pos in element:
            return element
        return None


class InstructionMappingElement:
    __slots__ = ("ins_addr", "posmap_pos")

    def __init__(self, ins_addr, posmap_pos):
        self.ins_addr: int = ins_addr
        self.posmap_pos: int = posmap_pos

    def __contains__(self, offset: int):
        return self.ins_addr == offset

    def __repr__(self):
        return f"<{self.ins_addr}: {self.posmap_pos}>"


class InstructionMapping:
    __slots__ = ("_insmap",)

    def __init__(self):
        self._insmap: SortedDict | dict[int, InstructionMappingElement] = SortedDict()

    def items(self):
        return self._insmap.items()

    def add_mapping(self, ins_addr, posmap_pos):
        if ins_addr in self._insmap:
            if posmap_pos <= self._insmap[ins_addr].posmap_pos:
                self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)
        else:
            self._insmap[ins_addr] = InstructionMappingElement(ins_addr, posmap_pos)

    def get_nearest_pos(self, ins_addr: int) -> int | None:
        try:
            pre_max = next(self._insmap.irange(maximum=ins_addr, reverse=True))
            pre_min = next(self._insmap.irange(minimum=ins_addr, reverse=False))
        except StopIteration:
            return None

        e1: InstructionMappingElement = self._insmap[pre_max]
        e2: InstructionMappingElement = self._insmap[pre_min]

        if abs(ins_addr - e1.ins_addr) <= abs(ins_addr - e2.ins_addr):
            return e1.posmap_pos
        return e2.posmap_pos


class BaseStructuredCodeGenerator:
    def __init__(self, flavor=None, notes=None, expr_comments=None, stmt_comments=None, const_formats=None):
        self.flavor = flavor
        self.text = None
        self.map_pos_to_node = None
        self.map_pos_to_addr = None
        self.map_addr_to_pos = None
        self.map_ast_to_pos: dict[SimVariable, set[PositionMappingElement]] | None = None
        self.notes = notes if notes is not None else {}
        self.expr_comments: dict[int, str] = expr_comments if expr_comments is not None else {}
        self.stmt_comments: dict[int, str] = stmt_comments if stmt_comments is not None else {}
        self.const_formats: dict[IdentType, dict[str, bool]] = const_formats if const_formats is not None else {}
        self.idx_counters: dict[str, count] = {}

    @staticmethod
    def adjust_mapping_positions(
        offset: int,
        pos_to_node: PositionMapping,
        pos_to_addr: PositionMapping,
        addr_to_pos: InstructionMapping,
    ) -> tuple[PositionMapping, PositionMapping, InstructionMapping]:
        """
        Adjust positions in the mappings to account for the notes that are prepended to the text.

        :param offset:       The length of the notes to prepend.
        :param pos_to_node:  The position to node mapping.
        :param pos_to_addr:  The position to address mapping.
        :param addr_to_pos:  The address to position mapping.
        :return:             Adjusted mappings.
        """
        new_pos_to_node = PositionMapping()
        new_pos_to_addr = PositionMapping()
        new_addr_to_pos = InstructionMapping()

        for pos, node in pos_to_node.items():
            new_pos_to_node.add_mapping(pos + offset, node.length, node.obj)

        for pos, node in pos_to_addr.items():
            new_pos_to_addr.add_mapping(pos + offset, node.length, node.obj)

        for addr, pos in addr_to_pos.items():
            new_addr_to_pos.add_mapping(addr, pos.posmap_pos + offset)

        return new_pos_to_node, new_pos_to_addr, new_addr_to_pos

    def reapply_options(self, options):
        pass

    def regenerate_text(self) -> None:
        pass

    def reload_variable_types(self) -> None:
        pass

    def next_idx(self, key: str) -> str:
        if key not in self.idx_counters:
            self.idx_counters[key] = count()
        return f"{key}_{next(self.idx_counters[key])}"

    def reset_idx_counters(self) -> None:
        self.idx_counters = {}
